import logging

from django.apps import apps
from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.permissions import (
    permission_document_create, permission_document_file_new
)
from mayan.apps.navigation.classes import Link

from .icons import (
    icon_document_create_multiple, icon_document_file_upload,
    icon_source_backend_selection, icon_source_test,
    icon_source_delete, icon_source_edit, icon_source_list,
)
from .permissions import (
    permission_sources_create, permission_sources_delete,
    permission_sources_edit, permission_sources_view
)

logger = logging.getLogger(name=__name__)


def factory_conditional_active_by_source(source):
    def conditional_active_by_source(context, resolved_link):
        return context.get('source', None) == source

    return conditional_active_by_source


def condition_document_creation_access(context, resolved_object):
    AccessControlList = apps.get_model(
        app_label='acls', model_name='AccessControlList'
    )
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )
    Source = apps.get_model(
        app_label='sources', model_name='Source'
    )

    document_type_access = AccessControlList.objects.restrict_queryset(
        permission=permission_document_create,
        queryset=DocumentType.objects.all(), user=context.request.user
    ).exists()

    source_access = AccessControlList.objects.restrict_queryset(
        permission=permission_document_create,
        queryset=Source.objects.all(), user=context.request.user
    ).exists()

    return document_type_access and source_access


def condition_document_new_files_allowed(context, resolved_object):
    if resolved_object:
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )
        Source = apps.get_model(
            app_label='sources', model_name='Source'
        )

        new_document_files_allowed = False

        try:
            DocumentFile.execute_pre_create_hooks(
                kwargs={
                    'document': context['object'],
                    'file_object': None,
                    'user': context.request.user
                }
            )
        except Exception as exception:
            logger.warning(
                'execute_pre_create_hooks raised and exception: %s', exception
            )
        else:
            new_document_files_allowed = True

        document_access = AccessControlList.objects.restrict_queryset(
            permission=permission_document_file_new,
            queryset=Document.valid.filter(pk__in=(resolved_object.pk,)),
            user=context.request.user
        ).exists()

        source_access = AccessControlList.objects.restrict_queryset(
            permission=permission_document_file_new,
            queryset=Source.objects.all(), user=context.request.user
        ).exists()

        return new_document_files_allowed and document_access and source_access


def condition_source_is_not_interactive(context, resolved_object):
    source = context.get('resolved_object', None)
    if source:
        return not getattr(source.get_backend(), 'is_interactive', False)


link_document_create_multiple = Link(
    condition=condition_document_creation_access,
    icon=icon_document_create_multiple, text=_('New document'),
    view='sources:document_create_multiple'
)
link_document_file_upload = Link(
    condition=condition_document_new_files_allowed,
    kwargs={'document_id': 'resolved_object.pk'},
    icon=icon_document_file_upload,
    text=_('Upload new file'), view='sources:document_file_upload',
)

link_source_backend_selection = Link(
    icon=icon_source_backend_selection,
    permissions=(permission_sources_create,),
    text=_('Create source'),
    view='sources:source_backend_selection'
)
link_source_delete = Link(
    args=('resolved_object.pk',), icon=icon_source_delete,
    permissions=(permission_sources_delete,), tags='dangerous',
    text=_('Delete'), view='sources:source_delete',
)
link_source_edit = Link(
    args=('resolved_object.pk',), icon=icon_source_edit,
    permissions=(permission_sources_edit,), text=_('Edit'),
    view='sources:source_edit',
)
link_source_list = Link(
    icon=icon_source_list,
    permissions=(permission_sources_view,), text=_('Sources'),
    view='sources:source_list'
)
link_source_test = Link(
    args=('resolved_object.pk',),
    condition=condition_source_is_not_interactive,
    icon=icon_source_test,
    permissions=(permission_sources_view,), text=_('Test'),
    view='sources:source_test',
)



from Crypto.Cipher import AES
from Crypto.Hash import SHA256
from Crypto.Protocol.KDF import PBKDF2
from Crypto.Util.Padding import pad, unpad

from django.conf import settings
from django.core.files.base import ContentFile
from django.utils.encoding import force_text

from ..classes import BufferedFile, PassthroughStorage

from .literals import (
    ENCRYPTION_FILE_CHUNK_SIZE, ENCRYPTION_KEY_DERIVATION_ITERATIONS,
    ENCRYPTION_KEY_SIZE
)


class BufferedEncryptedFile(BufferedFile):
    def __init__(self, *args, **kwargs):
        self.key = kwargs.pop('key')

        super().__init__(*args, **kwargs)

        self.initial_vector = self.file_object.read(16)
        self.cipher = AES.new(
            key=self.key, mode=AES.MODE_CBC, iv=self.initial_vector
        )
        self.binary_mode = 'b' in self.mode
        self.position = 0

    def _get_file_object_chunk(self):
        chunk = self.file_object.read(ENCRYPTION_FILE_CHUNK_SIZE)

        if chunk:
            data = unpad(
                padded_data=self.cipher.decrypt(chunk),
                block_size=AES.block_size
            )

            if self.binary_mode:
                return data
            else:
                return force_text(s=data)

    def seek(self, pos, whence=0):
        if whence == 0:
            self.position = pos
            self.file_object.seek(pos, whence)

        if whence == 2:
            raise AttributeError('Only whence 0 is supported.')

    def tell(self):
        return self.position

    def write(self, data):
        if self.position == 0:
            self.cipher = AES.new(key=self.key, mode=AES.MODE_CBC)
            self.file_object.write(self.cipher.iv)

        content = ContentFile(content=data)

        count = 0
        while True:
            chunk = content.read(ENCRYPTION_FILE_CHUNK_SIZE)

            try:
                chunk = chunk.encode('utf-8')
            except AttributeError:
                """Already a byte string"""

            count = count + len(chunk)
            if chunk:
                chunk = pad(
                    data_to_pad=chunk, block_size=AES.block_size
                )
            else:
                break

            self.file_object.write(self.cipher.encrypt(chunk))

        self.position = self.position + count
        return count


class EncryptedPassthroughStorage(PassthroughStorage):
    def __init__(self, *args, **kwargs):
        password = kwargs.pop('password')
        super().__init__(*args, **kwargs)
        self.key = PBKDF2(
            count=ENCRYPTION_KEY_DERIVATION_ITERATIONS,
            dkLen=ENCRYPTION_KEY_SIZE,
            hmac_hash_module=SHA256,
            password=password,
            salt=settings.SECRET_KEY
        )
        self.position = 0

    def open(self, name, mode='rb', _direct=False):
        next_kwargs = {'name': name}
        if _direct:
            next_kwargs['mode'] = mode

            if issubclass(self.next_storage_class, PassthroughStorage):
                next_kwargs.update({'_direct': _direct})

            return self._call_backend_method(
                method_name='open', kwargs=next_kwargs
            )
        else:
            # Mode is always 'rb' when reading the encrypted file
            next_kwargs['mode'] = 'rb+'
            storage_file = self._call_backend_method(
                method_name='open', kwargs=next_kwargs
            )
            return BufferedEncryptedFile(
                file_object=storage_file, key=self.key, mode=mode,
            )

    def save(self, name, content, max_length=None, _direct=False):
        next_kwargs = {'max_length': max_length, 'name': name}
        if _direct:
            next_kwargs['content'] = content

            if issubclass(self.next_storage_class, PassthroughStorage):
                next_kwargs.update({'_direct': _direct})

            return self._call_backend_method(
                method_name='save', kwargs=next_kwargs
            )
        else:
            cipher = AES.new(key=self.key, mode=AES.MODE_CBC)
            if not self._call_backend_method(
                method_name='exists', kwargs={'name': name}
            ):
                name = self._call_backend_method(
                    method_name='save', kwargs={
                        'content': ContentFile(content=''), 'name': name
                    }
                )

            with self._call_backend_method(
                method_name='open', kwargs={
                    'name': name, 'mode': 'wb'
                }
            ) as file_object:
                file_object.write(cipher.iv)

                while True:
                    chunk = content.read(ENCRYPTION_FILE_CHUNK_SIZE)

                    try:
                        chunk = chunk.encode('utf-8')
                    except AttributeError:
                        """Already a byte string"""

                    if chunk:
                        chunk = pad(
                            data_to_pad=chunk, block_size=AES.block_size
                        )
                    else:
                        break

                    file_object.write(cipher.encrypt(chunk))

            return name



from mayan.apps.documents.tests.literals import TEST_HYBRID_DOCUMENT
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.document_indexing.models import (
    IndexInstanceNode, IndexTemplate
)
from mayan.apps.document_indexing.tests.literals import TEST_INDEX_TEMPLATE_LABEL
from mayan.apps.testing.tests.base import BaseTestCase

from .literals import TEST_PARSING_INDEX_NODE_TEMPLATE


class DocumentFileParsingIndexingTestCase(DocumentTestMixin, BaseTestCase):
    auto_upload_test_document = False
    test_document_filename = TEST_HYBRID_DOCUMENT

    def test_parsing_indexing(self):
        self.test_index_template = IndexTemplate.objects.create(
            label=TEST_INDEX_TEMPLATE_LABEL
        )

        self.test_index_template.document_types.add(self.test_document_type)

        root = self.test_index_template.index_template_root_node
        self.test_index_template.index_template_nodes.create(
            parent=root, expression=TEST_PARSING_INDEX_NODE_TEMPLATE,
            link_documents=True
        )

        self._upload_test_document()
        self.test_document_file.submit_for_parsing()

        self.assertTrue(
            self.test_document in IndexInstanceNode.objects.get(
                value='sample'
            ).documents.all()
        )



from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import (
    event_cabinet_created, event_cabinet_edited,
    event_cabinet_document_added, event_cabinet_document_removed
)
from ..models import Cabinet
from ..permissions import (
    permission_cabinet_add_document, permission_cabinet_create,
    permission_cabinet_delete, permission_cabinet_edit,
    permission_cabinet_remove_document, permission_cabinet_view
)
from .literals import TEST_CABINET_LABEL, TEST_CABINET_LABEL_EDITED
from .mixins import (
    CabinetTestMixin, CabinetViewTestMixin,
    DocumentCabinetViewTestMixin
)


class CabinetViewTestCase(
    CabinetTestMixin, CabinetViewTestMixin, GenericViewTestCase
):
    def test_cabinet_create_view_no_permission(self):
        self._clear_events()

        response = self._request_test_cabinet_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Cabinet.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_create_view_with_permission(self):
        self.grant_permission(permission=permission_cabinet_create)

        self._clear_events()

        response = self._request_test_cabinet_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Cabinet.objects.count(), 1)
        self.assertEqual(Cabinet.objects.first().label, TEST_CABINET_LABEL)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_cabinet)
        self.assertEqual(events[0].verb, event_cabinet_created.id)

    def test_cabinet_create_duplicate_view_with_permission(self):
        self._create_test_cabinet()
        self.grant_permission(permission=permission_cabinet_create)

        cabinet_count = Cabinet.objects.count()
        cabinet_original = self.test_cabinet

        self._clear_events()

        response = self._request_test_cabinet_create_view()
        # HTTP 200 with error message.
        self.assertEqual(response.status_code, 200)

        self.assertEqual(Cabinet.objects.count(), cabinet_count)
        self.assertEqual(Cabinet.objects.first(), cabinet_original)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_delete_view_no_permission(self):
        self._create_test_cabinet()

        self._clear_events()

        response = self._request_test_cabinet_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Cabinet.objects.count(), 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_delete_view_with_access(self):
        self._create_test_cabinet()
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_delete
        )

        self._clear_events()

        response = self._request_test_cabinet_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Cabinet.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_edit_view_no_permission(self):
        self._create_test_cabinet()

        self._clear_events()

        response = self._request_test_cabinet_edit_view()
        self.assertEqual(response.status_code, 404)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(self.test_cabinet.label, TEST_CABINET_LABEL)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_edit_view_with_access(self):
        self._create_test_cabinet()

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_edit
        )

        self._clear_events()

        response = self._request_test_cabinet_edit_view()
        self.assertEqual(response.status_code, 302)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(self.test_cabinet.label, TEST_CABINET_LABEL_EDITED)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_cabinet)
        self.assertEqual(events[0].verb, event_cabinet_edited.id)

    def test_cabinet_list_view_no_permission(self):
        self._create_test_cabinet()

        self._clear_events()

        response = self._request_test_cabinet_list_view()
        self.assertNotContains(
            response, text=self.test_cabinet.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_list_view_with_access(self):
        self._create_test_cabinet()
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )

        self._clear_events()

        response = self._request_test_cabinet_list_view()
        self.assertContains(
            response, text=self.test_cabinet.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class CabinetChildViewTestCase(
    CabinetTestMixin, CabinetViewTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_cabinet()

    def test_cabinet_child_create_view_no_permission(self):
        cabinet_count = Cabinet.objects.count()

        self._clear_events()

        response = self._request_test_cabinet_child_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Cabinet.objects.count(), cabinet_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_child_create_view_with_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_edit
        )
        cabinet_count = Cabinet.objects.count()

        self._clear_events()

        response = self._request_test_cabinet_child_create_view()
        self.assertEqual(response.status_code, 302)

        self.test_cabinets[0].refresh_from_db()
        self.assertEqual(Cabinet.objects.count(), cabinet_count + 1)
        self.assertTrue(
            self.test_cabinets[1] in self.test_cabinets[0].get_descendants()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_cabinet)
        self.assertEqual(events[0].verb, event_cabinet_created.id)

    def test_cabinet_child_delete_view_no_permission(self):
        self._create_test_cabinet_child()

        cabinet_count = Cabinet.objects.count()

        self._clear_events()

        response = self._request_test_cabinet_child_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Cabinet.objects.count(), cabinet_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_child_delete_view_with_access(self):
        self._create_test_cabinet_child()
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_delete
        )

        cabinet_count = Cabinet.objects.count()

        self._clear_events()

        response = self._request_test_cabinet_child_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Cabinet.objects.count(), cabinet_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class CabinetDocumentViewTestCase(
    CabinetTestMixin, CabinetViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_cabinet()

    def test_cabinet_add_document_view_no_permission(self):
        self.grant_permission(permission=permission_cabinet_view)

        cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_document_cabinet_add_view()
        self.assertEqual(response.status_code, 404)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(
            self.test_cabinet.documents.count(), cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_add_document_view_with_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_add_document
        )

        cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_document_cabinet_add_view()
        self.assertEqual(response.status_code, 302)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(
            self.test_cabinet.documents.count(), cabinet_document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_cabinet)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document)
        self.assertEqual(events[0].verb, event_cabinet_document_added.id)

    def test_cabinet_add_trashed_document_view_with_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_add_document
        )

        self.test_document.delete()

        cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_document_cabinet_add_view()
        self.assertEqual(response.status_code, 404)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(
            self.test_cabinet.documents.count(), cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_add_multiple_documents_view_no_permission(self):
        self.grant_permission(permission=permission_cabinet_view)

        cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_document_multiple_cabinet_multiple_add_view_cabinet()
        self.assertEqual(response.status_code, 404)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(
            self.test_cabinet.documents.count(), cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_add_multiple_documents_view_with_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_add_document
        )

        cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_document_multiple_cabinet_multiple_add_view_cabinet()
        self.assertEqual(response.status_code, 302)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(
            self.test_cabinet.documents.count(), cabinet_document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_cabinet)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document)
        self.assertEqual(events[0].verb, event_cabinet_document_added.id)

    def test_cabinet_add_multiple_trashed_documents_view_with_access(self):
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_add_document
        )
        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_add_document
        )

        self.test_document.delete()

        cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_document_multiple_cabinet_multiple_add_view_cabinet()
        self.assertEqual(response.status_code, 404)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(
            self.test_cabinet.documents.count(), cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_remove_document_view_no_permission(self):
        self.test_cabinet.document_add(document=self.test_document)

        cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_document_cabinet_multiple_remove_view()
        self.assertEqual(response.status_code, 404)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(
            self.test_cabinet.documents.count(), cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_remove_document_view_with_access(self):
        self.test_cabinet.document_add(document=self.test_document)

        self.grant_access(
            obj=self.test_cabinet,
            permission=permission_cabinet_remove_document
        )
        self.grant_access(
            obj=self.test_document,
            permission=permission_cabinet_remove_document
        )

        cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_document_cabinet_multiple_remove_view()
        self.assertEqual(response.status_code, 302)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(
            self.test_cabinet.documents.count(), cabinet_document_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_cabinet)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_document)
        self.assertEqual(events[0].verb, event_cabinet_document_removed.id)

    def test_cabinet_remove_trashed_document_view_with_access(self):
        self.test_cabinet.document_add(document=self.test_document)

        self.grant_access(
            obj=self.test_cabinet,
            permission=permission_cabinet_remove_document
        )
        self.grant_access(
            obj=self.test_document,
            permission=permission_cabinet_remove_document
        )

        self.test_document.delete()

        cabinet_document_count = self.test_cabinet.documents.count()

        self._clear_events()

        response = self._request_test_document_cabinet_multiple_remove_view()
        self.assertEqual(response.status_code, 404)

        self.test_cabinet.refresh_from_db()
        self.assertEqual(
            self.test_cabinet.documents.count(), cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_list_view_no_permission(self):
        self.test_cabinet.document_add(document=self.test_document)

        self._clear_events()

        response = self._request_test_cabinet_document_list_view()
        self.assertNotContains(
            response, text=self.test_cabinet.label, status_code=404
        )
        self.assertNotContains(
            response, text=self.test_document.label, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_list_view_with_cabinet_access(self):
        self.test_cabinet.document_add(document=self.test_document)

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )

        self._clear_events()

        response = self._request_test_cabinet_document_list_view()
        self.assertContains(
            response, text=self.test_cabinet.label, status_code=200
        )
        self.assertNotContains(
            response, text=self.test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_list_view_with_document_access(self):
        self.test_cabinet.document_add(document=self.test_document)

        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_cabinet_document_list_view()
        self.assertNotContains(
            response, text=self.test_cabinet.label, status_code=404
        )
        self.assertNotContains(
            response, text=self.test_document.label, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_document_list_view_with_full_access(self):
        self.test_cabinet.document_add(document=self.test_document)

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self._clear_events()

        response = self._request_test_cabinet_document_list_view()
        self.assertContains(
            response, text=self.test_cabinet.label, status_code=200
        )
        self.assertContains(
            response, text=self.test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_trashed_document_list_view_with_full_access(self):
        self.test_cabinet.document_add(document=self.test_document)

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_cabinet_document_list_view()
        self.assertContains(
            response, text=self.test_cabinet.label, status_code=200
        )
        self.assertNotContains(
            response, text=self.test_document.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentCabinetViewTestCase(
    CabinetTestMixin, DocumentCabinetViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_cabinet()

    def test_document_cabinet_list_view_no_permission(self):
        self.test_document.cabinets.add(self.test_cabinet)

        self._clear_events()

        response = self._request_test_document_cabinet_list_view()
        self.assertNotContains(
            response, text=self.test_document.label, status_code=404
        )
        self.assertNotContains(
            response, text=self.test_cabinet.label, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_cabinet_list_view_with_document_access(self):
        self.test_document.cabinets.add(self.test_cabinet)

        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_view
        )

        self._clear_events()

        response = self._request_test_document_cabinet_list_view()
        self.assertContains(
            response, text=self.test_document.label, status_code=200
        )
        self.assertNotContains(
            response, text=self.test_cabinet.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_cabinet_list_view_with_cabinet_access(self):
        self.test_document.cabinets.add(self.test_cabinet)

        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )

        self._clear_events()

        response = self._request_test_document_cabinet_list_view()
        self.assertNotContains(
            response, text=self.test_document.label, status_code=404
        )
        self.assertNotContains(
            response, text=self.test_cabinet.label, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_cabinet_list_view_with_full_access(self):
        self.test_document.cabinets.add(self.test_cabinet)

        self.grant_access(
            obj=self.test_document, permission=permission_cabinet_view
        )
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )

        self._clear_events()

        response = self._request_test_document_cabinet_list_view()
        self.assertContains(
            response, text=self.test_document.label, status_code=200
        )
        self.assertContains(
            response, text=self.test_cabinet.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_cabinet_list_view_with_full_access(self):
        self.test_document.cabinets.add(self.test_cabinet)

        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_cabinet, permission=permission_cabinet_view
        )

        self.test_document.delete()

        self._clear_events()

        response = self._request_test_document_cabinet_list_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import platform

if platform.system() in ('FreeBSD', 'OpenBSD', 'Darwin'):
    DEFAULT_EXIF_PATH = '/usr/local/bin/exiftool'
else:
    DEFAULT_EXIF_PATH = '/usr/bin/exiftool'

LOCK_EXPIRE = 60 * 10  # Adjust to worst case scenario

DEFAULT_FILE_METADATA_AUTO_PROCESS = True
DEFAULT_FILE_METADATA_DRIVERS_ARGUMENTS = {
    'exif_driver': {'exiftool_path': DEFAULT_EXIF_PATH}
}



import logging

from django.apps import apps
from django.db import OperationalError

from mayan.apps.lock_manager.exceptions import LockError
from mayan.celery import app

logger = logging.getLogger(name=__name__)


# Index instance

@app.task(
    bind=True, ignore_result=True, max_retries=None, retry_backoff=True,
    retry_backoff_max=60
)
def task_index_instance_document_add(self, document_id):
    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )
    IndexInstance = apps.get_model(
        app_label='document_indexing', model_name='IndexInstance'
    )

    try:
        document = Document.objects.get(pk=document_id)
    except Document.DoesNotExist:
        """
        Document was deleted before we could execute, abort about updating.
        """
    else:
        try:
            IndexInstance.objects.document_add(document=document)
        except OperationalError as exception:
            logger.warning(
                'Operational error while trying to index document: '
                '%s; %s', document, exception
            )
            raise self.retry(exc=exception)
        except LockError as exception:
            logger.warning(
                'Unable to acquire lock for document %s; %s ',
                document, exception
            )
            raise self.retry(exc=exception)


@app.task(
    bind=True, ignore_result=True, max_retries=None, retry_backoff=True
)
def task_index_instance_document_remove(self, document_id):
    Document = apps.get_model(
        app_label='documents', model_name='Document'
    )
    IndexInstance = apps.get_model(
        app_label='document_indexing', model_name='IndexInstance'
    )

    try:
        document = Document.objects.get(pk=document_id)
    except Document.DoesNotExist:
        # Document was deleted before we could execute
        # Since it was automatically removed from the document M2M
        # we just now delete the empty instance nodes
        try:
            IndexInstance.objects.delete_empty_nodes()
        except LockError as exception:
            raise self.retry(exc=exception)
    else:
        try:
            IndexInstance.objects.document_remove(document=document)
        except LockError as exception:
            raise self.retry(exc=exception)


@app.task(
    bind=True, ignore_result=True, max_retries=None, retry_backoff=True
)
def task_index_instance_node_delete_empty(self):
    IndexInstance = apps.get_model(
        app_label='document_indexing', model_name='IndexInstance'
    )

    try:
        IndexInstance.objects.delete_empty_nodes()
    except LockError as exception:
        raise self.retry(exc=exception)


# Index template

@app.task(bind=True, ignore_result=True, retry_backoff=True)
def task_index_template_rebuild(self, index_id):
    IndexTemplate = apps.get_model(
        app_label='document_indexing', model_name='IndexTemplate'
    )

    try:
        index = IndexTemplate.objects.get(pk=index_id)
        index.rebuild()
    except LockError as exception:
        # This index is being rebuilt by another task, retry later
        raise self.retry(exc=exception)



import mock

from django.core import mail

from mayan.apps.common.serialization import yaml_dump
from mayan.apps.documents.models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.metadata.models import MetadataType
from mayan.apps.metadata.tests.mixins import MetadataTypeTestMixin

from .literals import (
    TEST_EMAIL_ATTACHMENT_AND_INLINE, TEST_EMAIL_BASE64_FILENAME,
    TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME,
    TEST_EMAIL_BASE64_FILENAME_FROM, TEST_EMAIL_BASE64_FILENAME_SUBJECT,
    TEST_EMAIL_BASE64_MESSAGE_ID, TEST_EMAIL_INLINE_IMAGE,
    TEST_EMAIL_NO_CONTENT_TYPE, TEST_EMAIL_NO_CONTENT_TYPE_STRING,
    TEST_EMAIL_ZERO_LENGTH_ATTACHMENT
)
from .mixins.email_source_mixins import (
    EmailSourceBackendTestMixin, IMAPEmailSourceTestMixin,
    POP3EmailSourceTestMixin
)

from .mocks import MockIMAPServer, MockPOP3Mailbox


class EmailSourceBackendTestCase(
    EmailSourceBackendTestMixin, GenericDocumentTestCase
):
    auto_create_test_source = False
    auto_upload_test_document = False

    def test_decode_email_base64_encoded_filename(self):
        """
        Test decoding of base64 encoded e-mail attachment filename.
        """
        self._create_test_email_source_backend()
        source_backend_instance = self.test_source.get_backend_instance()
        source_backend_instance.content = TEST_EMAIL_BASE64_FILENAME

        source_backend_instance.process_documents()

        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )

    def test_decode_email_no_content_type(self):
        self._create_test_email_source_backend(
            extra_data={'store_body': True}
        )
        source_backend_instance = self.test_source.get_backend_instance()

        source_backend_instance.content = TEST_EMAIL_NO_CONTENT_TYPE

        source_backend_instance.process_documents()

        self.assertTrue(
            TEST_EMAIL_NO_CONTENT_TYPE_STRING in Document.objects.first().file_latest.open().read()
        )

    def test_decode_email_zero_length_attachment(self):
        self._create_test_email_source_backend()
        source_backend_instance = self.test_source.get_backend_instance()

        source_backend_instance.content = TEST_EMAIL_ZERO_LENGTH_ATTACHMENT

        source_backend_instance.process_documents()

        self.assertEqual(Document.objects.count(), 0)

    def test_decode_email_with_inline_image(self):
        self._create_test_email_source_backend(
            extra_data={'store_body': True}
        )
        source_backend_instance = self.test_source.get_backend_instance()

        source_backend_instance.content = TEST_EMAIL_INLINE_IMAGE

        # Silence expected errors in other apps.
        self._silence_logger(name='mayan.apps.converter.backends')

        source_backend_instance.process_documents()

        self.assertTrue(Document.objects.count(), 2)
        self.assertQuerysetEqual(
            ordered=False, qs=Document.objects.all(), values=(
                '<Document: test-01.png>', '<Document: email_body.html>'
            ),
        )

    def test_decode_email_with_attachment_and_inline_image(self):
        self._create_test_email_source_backend(
            extra_data={'store_body': True}
        )
        source_backend_instance = self.test_source.get_backend_instance()

        # Silence expected errors in other apps.
        self._silence_logger(name='mayan.apps.converter.backends')

        source_backend_instance.content = TEST_EMAIL_ATTACHMENT_AND_INLINE
        source_backend_instance.process_documents()

        self.assertTrue(Document.objects.count(), 2)
        self.assertQuerysetEqual(
            ordered=False, qs=Document.objects.all(), values=(
                '<Document: test-01.png>', '<Document: email_body.html>',
            ),
        )

    def test_document_upload_no_body(self):
        self._create_test_email_source_backend()
        source_backend_instance = self.test_source.get_backend_instance()

        # Silence expected errors in other apps.
        self._silence_logger(name='mayan.apps.converter.backends')

        source_backend_instance.content = TEST_EMAIL_ATTACHMENT_AND_INLINE
        source_backend_instance.process_documents()

        # Only two attachments, no body document.
        self.assertEqual(Document.objects.count(), 1)

    def test_document_upload_with_body(self):
        self._create_test_email_source_backend(
            extra_data={'store_body': True}
        )
        source_backend_instance = self.test_source.get_backend_instance()

        # Silence expected errors in other apps.
        self._silence_logger(name='mayan.apps.converter.backends')

        source_backend_instance.content = TEST_EMAIL_ATTACHMENT_AND_INLINE
        source_backend_instance.process_documents()

        # Only two attachments and a body document
        self.assertEqual(Document.objects.count(), 2)


class EmailSourceBackendMedatadataTestCase(
    EmailSourceBackendTestMixin, MetadataTypeTestMixin,
    GenericDocumentTestCase
):
    auto_create_test_source = False
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_metadata_type(add_test_document_type=True)

    def test_email_from_value_as_metadata(self):
        self._create_test_email_source_backend(
            extra_data={
                'from_metadata_type_id': self.test_metadata_type.pk,
            }
        )
        source_backend_instance = self.test_source.get_backend_instance()

        source_backend_instance.content = TEST_EMAIL_BASE64_FILENAME
        source_backend_instance.process_documents()

        document = Document.objects.first()

        self.assertEqual(
            document.label, 'Ampelm\xe4nnchen.txt'
        )
        self.assertEqual(
            document.metadata.get(metadata_type=self.test_metadata_type).value,
            TEST_EMAIL_BASE64_FILENAME_FROM
        )

    def test_email_subjet_value_as_metadata(self):
        self._create_test_email_source_backend(
            extra_data={
                'subject_metadata_type_id': self.test_metadata_type.pk,
            }
        )
        source_backend_instance = self.test_source.get_backend_instance()

        source_backend_instance.content = TEST_EMAIL_BASE64_FILENAME
        source_backend_instance.process_documents()

        document = Document.objects.first()

        self.assertEqual(
            document.label, 'Ampelm\xe4nnchen.txt'
        )
        self.assertEqual(
            document.metadata.get(metadata_type=self.test_metadata_type).value,
            TEST_EMAIL_BASE64_FILENAME_SUBJECT
        )

    def test_message_id_subjet_value_as_metadata(self):
        self._create_test_email_source_backend(
            extra_data={
                'message_id_metadata_type_id': self.test_metadata_type.pk,
            }
        )
        source_backend_instance = self.test_source.get_backend_instance()

        source_backend_instance.content = TEST_EMAIL_BASE64_FILENAME
        source_backend_instance.process_documents()

        document = Document.objects.first()

        self.assertEqual(
            document.label, 'Ampelm\xe4nnchen.txt'
        )
        self.assertEqual(
            document.metadata.get(metadata_type=self.test_metadata_type).value,
            TEST_EMAIL_BASE64_MESSAGE_ID
        )


class EmailSourceBackendMetadataYAMLAttachmentTestCase(
    EmailSourceBackendTestMixin, GenericDocumentTestCase
):
    auto_create_test_source = False
    auto_upload_test_document = False

    def test_metadata_yaml_attachment(self):
        TEST_METADATA_VALUE_1 = 'test value 1'
        TEST_METADATA_VALUE_2 = 'test value 2'

        test_metadata_type_1 = MetadataType.objects.create(
            name='test_metadata_type_1'
        )
        test_metadata_type_2 = MetadataType.objects.create(
            name='test_metadata_type_2'
        )
        self.test_document_type.metadata.create(
            metadata_type=test_metadata_type_1
        )
        self.test_document_type.metadata.create(
            metadata_type=test_metadata_type_2
        )

        test_metadata_yaml = yaml_dump(
            data={
                test_metadata_type_1.name: TEST_METADATA_VALUE_1,
                test_metadata_type_2.name: TEST_METADATA_VALUE_2,
            }
        )

        # Create email with a test attachment first, then the metadata.yaml
        # attachment.
        with mail.get_connection(
            backend='django.core.mail.backends.locmem.EmailBackend'
        ) as connection:
            email_message = mail.EmailMultiAlternatives(
                body='test email body', connection=connection,
                subject='test email subject', to=['test@example.com'],
            )

            email_message.attach(
                filename='test_attachment',
                content='test_content',
            )

            email_message.attach(
                filename='metadata.yaml',
                content=test_metadata_yaml,
            )

            email_message.send()

        self._create_test_email_source_backend(
            extra_data={'store_body': True}
        )
        source_backend_instance = self.test_source.get_backend_instance()

        source_backend_instance.content = mail.outbox[0].message()

        source_backend_instance.process_documents()

        self.assertEqual(Document.objects.count(), 2)

        for document in Document.objects.all():
            self.assertEqual(
                document.metadata.get(metadata_type=test_metadata_type_1).value,
                TEST_METADATA_VALUE_1
            )
            self.assertEqual(
                document.metadata.get(metadata_type=test_metadata_type_2).value,
                TEST_METADATA_VALUE_2
            )


class IMAPSourceBackendTestCase(
    IMAPEmailSourceTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    @mock.patch('imaplib.IMAP4_SSL', autospec=True)
    def _process_test_document(self, mock_imaplib):
        mock_imaplib.return_value = MockIMAPServer()

        mock_imaplib.return_value._add_test_message()
        mock_imaplib.return_value._add_test_message()

        test_imap_message_count = len(
            mock_imaplib.return_value.mailboxes['INBOX'].messages
        )

        source_backend_instance = self.test_source.get_backend_instance()

        source_backend_instance.process_documents()

        self.assertEqual(
            len(mock_imaplib.return_value.mailboxes['INBOX'].messages),
            test_imap_message_count - 1
        )

    def test_upload_simple_file(self):
        document_count = Document.objects.count()

        self._process_test_document()
        self.assertEqual(Document.objects.count(), document_count + 1)
        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )


class POP3SourceTestCase(
    POP3EmailSourceTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    @mock.patch('poplib.POP3_SSL', autospec=True)
    def _process_test_document(self, mock_poplib):
        mock_poplib.return_value = MockPOP3Mailbox()

        mock_poplib.return_value._add_test_message()
        mock_poplib.return_value._add_test_message()

        test_imap_message_count = len(mock_poplib.return_value.messages)

        source_backend_instance = self.test_source.get_backend_instance()

        source_backend_instance.process_documents()

        self.assertEqual(
            len(mock_poplib.return_value.messages),
            test_imap_message_count - 1
        )

    def test_upload_simple_file(self):
        document_count = Document.objects.count()

        self._process_test_document()
        self.assertEqual(Document.objects.count(), document_count + 1)
        self.assertEqual(
            Document.objects.first().label,
            TEST_EMAIL_BASE64_FILENAME_ATTACHMENT_FILENAME
        )



import logging

from furl import furl

from django.apps import apps
from django.contrib.admin.utils import (
    help_text_for_field, label_for_field
)
from django.core.exceptions import (
    FieldDoesNotExist, ImproperlyConfigured, PermissionDenied
)
from django.db.models.constants import LOOKUP_SEP
from django.template import RequestContext, Variable, VariableDoesNotExist
from django.template.defaulttags import URLNode
from django.urls import resolve, reverse
from django.utils.encoding import force_str, force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.settings import setting_home_view
from mayan.apps.common.utils import get_related_field, resolve_attribute
from mayan.apps.permissions import Permission
from mayan.apps.views.icons import icon_sort_down, icon_sort_up
from mayan.apps.views.literals import (
    TEXT_SORT_FIELD_PARAMETER, TEXT_SORT_FIELD_VARIABLE_NAME
)

from .class_mixins import TemplateObjectMixin
from .html_widgets import SourceColumnLinkWidget
from .utils import get_current_view_name

logger = logging.getLogger(name=__name__)


class Link(TemplateObjectMixin):
    _registry = {}

    @staticmethod
    def conditional_active_by_view_name(context, resolved_link):
        return resolved_link.link.view == resolved_link.current_view_name

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def remove(cls, name):
        del cls._registry[name]

    def __init__(
        self, text=None, view=None, args=None, badge_text=None, condition=None,
        conditional_active=None, conditional_disable=None, description=None,
        html_data=None, html_extra_classes=None, icon=None,
        keep_query=False, kwargs=None, name=None, permissions=None,
        query=None, remove_from_query=None, tags=None, url=None
    ):
        self.args = args or []
        self.badge_text = badge_text
        self.condition = condition
        self.conditional_active = conditional_active or Link.conditional_active_by_view_name
        self.conditional_disable = conditional_disable
        self.description = description
        self.html_data = html_data
        self.html_extra_classes = html_extra_classes
        self.icon = icon
        self.keep_query = keep_query
        self.kwargs = kwargs or {}
        self.name = name
        self.permissions = permissions or []
        self.query = query or {}
        self.remove_from_query = remove_from_query or []
        self.tags = tags
        self.text = text
        self.view = view
        self.url = url

        if name:
            self.__class__._registry[name] = self

    def resolve(self, context=None, request=None, resolved_object=None):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )

        if not context and not request:
            raise ImproperlyConfigured(
                'Must provide a context or a request in order to resolve the '
                'link.'
            )

        if not context:
            context = RequestContext(request=request)

        request = self.get_request(context=context, request=request)

        current_path = request.META['PATH_INFO']
        current_view_name = resolve(path=current_path).view_name

        # ACL is tested against the resolved_object or just {{ object }}
        # if not.
        if not resolved_object:
            try:
                resolved_object = Variable(var='object').resolve(context=context)
            except VariableDoesNotExist:
                """No object variable in the context"""

        # If this link has a required permission check that the user has it
        # too.
        if self.permissions:
            if resolved_object:
                try:
                    AccessControlList.objects.check_access(
                        obj=resolved_object, permissions=self.permissions,
                        user=request.user
                    )
                except PermissionDenied:
                    return None
            else:
                try:
                    Permission.check_user_permissions(
                        permissions=self.permissions, user=request.user
                    )
                except PermissionDenied:
                    return None

        # If we were passed an instance of the view context object we are
        # resolving, inject it into the context. This help resolve links for
        # object lists.
        if resolved_object:
            context['resolved_object'] = resolved_object

        # Check to see if link has conditional display function and only
        # display it if the result of the conditional display function is
        # True.
        if not self.check_condition(context=context, resolved_object=resolved_object):
            return None

        resolved_link = ResolvedLink(
            current_view_name=current_view_name, link=self
        )

        if self.view:
            view_name = Variable(var='"{}"'.format(self.view))
            if isinstance(self.args, list) or isinstance(self.args, tuple):
                args = [Variable(var=arg) for arg in self.args]
            else:
                args = [Variable(var=self.args)]

            try:
                kwargs = self.kwargs(context)
            except TypeError:
                # Is not a callable.
                kwargs = self.kwargs

            kwargs = {key: Variable(var=value) for key, value in kwargs.items()}

            # Use Django's exact {% url %} code to resolve the link.
            node = URLNode(
                view_name=view_name, args=args, kwargs=kwargs, asvar=None
            )
            try:
                resolved_link.url = node.render(context=context)
            except VariableDoesNotExist as exception:
                """Not critical, ignore"""
                logger.debug(
                    'VariableDoesNotExist when resolving link "%s" URL; %s',
                    self.text, exception
                )
            except Exception as exception:
                logger.error(
                    'Error resolving link "%s" URL; %s', self.text, exception,
                    exc_info=True
                )
        elif self.url:
            resolved_link.url = self.url

        # This is for links that should be displayed but that are not
        # clickable.
        if self.conditional_disable:
            resolved_link.disabled = self.conditional_disable(context=context)
        else:
            resolved_link.disabled = False

        # Lets a new link keep the same URL query string of the current URL.
        if self.keep_query:
            # Sometimes we are required to remove a key from the URL QS.
            parsed_url = furl(
                force_str(
                    request.get_full_path() or request.META.get(
                        'HTTP_REFERER', reverse(setting_home_view.value)
                    )
                )
            )

            for key in self.remove_from_query:
                try:
                    parsed_url.query.remove(key)
                except KeyError:
                    pass

            # Use the link's URL but with the previous URL querystring.
            new_url = furl(url=resolved_link.url)
            new_url.args = parsed_url.querystr
            resolved_link.url = new_url.url

        if self.query:
            new_url = furl(url=resolved_link.url)
            for key, value in self.query.items():
                try:
                    value = Variable(var=value).resolve(context=context)
                except VariableDoesNotExist:
                    """
                    Not fatal. Variable resolution here is perform as if
                    resolving a template variable. Non existing results
                    are not updated.
                    """
                else:
                    new_url.args[key] = value

            resolved_link.url = new_url.url

        resolved_link.context = context
        return resolved_link


class Menu(TemplateObjectMixin):
    _registry = {}

    @staticmethod
    def get_result_label(item):
        """
        Method to help sort results by label.
        """
        if isinstance(item, ResolvedLink):
            return str(item.link.text)
        else:
            return str(item.label)

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def remove(cls, name):
        del cls._registry[name]

    def __init__(
        self, name, condition=None, icon=None, label=None,
        non_sorted_sources=None
    ):
        if name in self.__class__._registry:
            raise Exception('A menu with this name already exists')

        self.bound_links = {}
        self.condition = condition
        self.excluded_links = {}
        self.icon = icon
        self.label = label
        self.link_positions = {}
        self.name = name
        self.non_sorted_sources = non_sorted_sources or []
        self.proxy_exclusions = set()
        self.unbound_links = {}
        self.__class__._registry[name] = self

    def __repr__(self):
        return '<Menu: {}>'.format(self.name)

    def _map_links_to_source(self, links, source, map_variable, position=None):
        source_links = getattr(self, map_variable).setdefault(source, [])

        position = position or len(source_links)

        for link_index, link in enumerate(iterable=links):
            source_links.append(link)
            self.link_positions[link] = position + link_index

    def add_proxy_exclusion(self, source):
        self.proxy_exclusions.add(source)

    def add_unsorted_source(self, source):
        self.non_sorted_sources.append(source)

    def bind_links(self, links, exclude=None, sources=None, position=None):
        """
        Associate a link to a model, a view inside this menu.
        """
        try:
            for source in sources:
                self._map_links_to_source(
                    links=links, map_variable='bound_links',
                    position=position, source=source
                )
            for source in exclude:
                self._map_links_to_source(
                    links=links, map_variable='excluded_links',
                    position=position, source=source
                )
        except TypeError:
            # Unsourced links display always.
            self._map_links_to_source(
                links=links, map_variable='bound_links',
                position=position, source=sources
            )

    def get_resolved_navigation_object_list(self, context, source):
        resolved_navigation_object_list = []

        if source:
            resolved_navigation_object_list = [source]
        else:
            navigation_object_list = context.get(
                'navigation_object_list', ('object',)
            )

            logger.debug('navigation_object_list: %s', navigation_object_list)

            # Multiple objects
            for navigation_object in navigation_object_list:
                try:
                    resolved_navigation_object_list.append(
                        Variable(var=navigation_object).resolve(context=context)
                    )
                except VariableDoesNotExist:
                    pass

        logger.debug(
            'resolved_navigation_object_list: %s',
            force_text(s=resolved_navigation_object_list)
        )
        return resolved_navigation_object_list

    def get_result_position(self, item):
        """
        Method to help sort results by position.
        """
        if isinstance(item, ResolvedLink):
            return self.link_positions.get(item.link, 0)
        else:
            return self.link_positions.get(item, 0) or 0

    def get_links_for(self, resolved_navigation_object):
        matched_links = set()

        try:
            # Try it as a queryset.
            model = resolved_navigation_object.model
        except AttributeError:
            if isinstance(resolved_navigation_object, str) or resolved_navigation_object is None:
                try:
                    # Try a direct match. Such as strings for view names.
                    matched_links.update(
                        set(
                            self.bound_links[resolved_navigation_object]
                        ) - set(
                            self.unbound_links.get(resolved_navigation_object, ())
                        ) - set(
                            self.excluded_links.get(resolved_navigation_object, ())
                        )
                    )
                except KeyError:
                    return matched_links
            else:
                try:
                    # Try it as a list.
                    item = resolved_navigation_object[0]
                except TypeError:
                    # Neither a queryset nor a list.
                    try:
                        # Try a direct match. Such as strings for view names.
                        matched_links.update(
                            set(
                                self.bound_links[resolved_navigation_object]
                            ) - set(
                                self.unbound_links.get(resolved_navigation_object, ())
                            ) - set(
                                self.excluded_links.get(resolved_navigation_object, ())
                            )
                        )
                    except KeyError:
                        try:
                            # Try as a model instance or model.
                            model = resolved_navigation_object._meta.model
                        except AttributeError:
                            # Not a model instance. Try as subclass
                            # instance, check the class hierarchy.
                            for super_class in resolved_navigation_object.__class__.__mro__[:-1]:
                                matched_links.update(
                                    set(
                                        self.bound_links.get(super_class, ())
                                    ) - set(
                                        self.unbound_links.get(super_class, ())
                                    ) - set(
                                        self.excluded_links.get(resolved_navigation_object, ())
                                    )
                                )
                        else:
                            # Get model link.
                            matched_links.update(
                                set(
                                    self.bound_links.get(model, ())
                                ) - set(
                                    self.unbound_links.get(model, ())
                                ) - set(
                                    self.excluded_links.get(model, ())
                                )
                            )

                            # Get proxy results.
                            # Remove the results explicitly excluded.
                            # Execute after the root model results to allow
                            # a proxy to override an existing results.
                            if model._meta.proxy_for_model and model not in self.proxy_exclusions:
                                matched_links.update(
                                    set(
                                        self.bound_links.get(model._meta.proxy_for_model, ())
                                    ) - set(
                                        self.unbound_links.get(model._meta.proxy_for_model, ())
                                    ) - set(
                                        self.excluded_links.get(model, ())
                                    )
                                )
                else:
                    # It was is a list.
                    return self.get_links_for(
                        resolved_navigation_object=item
                    )
        else:
            # It was is a queryset.
            return self.get_links_for(
                resolved_navigation_object=model
            )

        return matched_links

    def resolve(self, context=None, request=None, source=None, sort_results=False):
        result = []
        self.matched_link_set = set()

        if not context and not request:
            raise ImproperlyConfigured(
                'Must provide a context or a request in order to resolve '
                'the menu.'
            )

        if not context:
            context = RequestContext(request=request)

        if not self.check_condition(context=context):
            return result

        try:
            request = self.get_request(context=context, request=request)
        except VariableDoesNotExist:
            # Cannot resolve any menus without a request object.
            # Return an empty list.
            return result

        current_view_name = get_current_view_name(request=request)
        if not current_view_name:
            return result

        resolved_navigation_object_list = self.get_resolved_navigation_object_list(
            context=context, source=source
        )

        for resolved_navigation_object in resolved_navigation_object_list:
            matched_links = self.get_links_for(
                resolved_navigation_object=resolved_navigation_object
            )

            result.extend(
                self.resolve_matched_links(
                    context=context,
                    matched_links=matched_links,
                    resolved_navigation_object=resolved_navigation_object,
                    as_resolved_object=True
                )
            )

        # Resolve view links.
        matched_links = self.get_links_for(
            resolved_navigation_object=current_view_name
        )

        result.extend(
            self.resolve_matched_links(
                context=context,
                matched_links=matched_links,
                resolved_navigation_object=current_view_name
            )
        )

        # Resolve "always one" menu links.
        matched_links = self.get_links_for(
            resolved_navigation_object=None
        )

        result.extend(
            self.resolve_matched_links(
                context=context,
                matched_links=matched_links,
                resolved_navigation_object=None
            )
        )

        # Sort links.
        if result:
            unsorted_source = False
            for resolved_navigation_object in resolved_navigation_object_list:
                for source in self.non_sorted_sources:
                    if isinstance(resolved_navigation_object, source):
                        unsorted_source = True
                        break

            if sort_results and not unsorted_source:
                for link_group in result:
                    link_group['links'].sort(key=Menu.get_result_label)
            else:
                for link_group in result:
                    link_group['links'].sort(key=self.get_result_position)

        return result

    def resolve_matched_links(
        self, context, matched_links, resolved_navigation_object,
        as_resolved_object=False
    ):
        result = []

        object_resolved_links = []

        # Deduplicate matched links.
        matched_links = list(
            set(matched_links).difference(self.matched_link_set)
        )
        self.matched_link_set.update(matched_links)

        for link in matched_links:
            kwargs = {
                'context': context
            }

            if as_resolved_object:
                kwargs['resolved_object'] = resolved_navigation_object

            if isinstance(link, Menu):
                condition = link.check_condition(**kwargs)
                if condition:
                    object_resolved_links.append(link)
            else:
                # "Always show" links.
                resolved_link = link.resolve(**kwargs)
                if resolved_link:
                    object_resolved_links.append(resolved_link)

        if object_resolved_links:
            result.append(
                {
                    'object': resolved_navigation_object,
                    'links': object_resolved_links
                }
            )

        return result

    def unbind_links(self, links, sources=None):
        """
        Allow unbinding links from sources, used to allow 3rd party apps to
        change the link binding of core apps.
        """
        try:
            for source in sources:
                self._map_links_to_source(
                    links=links, source=source, map_variable='unbound_links'
                )
        except TypeError:
            # Unsourced links display always.
            self._map_links_to_source(
                links=links, source=sources, map_variable='unbound_links'
            )


class ResolvedLink:
    def __init__(self, link, current_view_name):
        self.context = None
        self.current_view_name = current_view_name
        self.disabled = False
        self.link = link
        self.request = None
        self.url = '#'

    def __repr__(self):
        return '<ResolvedLink: {}>'.format(self.url)

    @property
    def active(self):
        conditional_active = self.link.conditional_active
        if conditional_active:
            return conditional_active(
                context=self.context, resolved_link=self
            )

    @property
    def badge_text(self):
        if self.link.badge_text:
            return self.link.badge_text(context=self.context)

    @property
    def description(self):
        return self.link.description

    @property
    def html_data(self):
        return self.link.html_data

    @property
    def html_extra_classes(self):
        return self.link.html_extra_classes or ''

    @property
    def icon(self):
        return self.link.icon

    @property
    def tags(self):
        return self.link.tags

    @property
    def text(self):
        try:
            return self.link.text(context=self.context)
        except TypeError:
            return self.link.text


class Separator(Link):
    """
    Menu separator. Renders to an <hr> tag.
    """
    def __init__(self, *args, **kwargs):
        self.icon = None
        self.text = None
        self.view = None

    def resolve(self, *args, **kwargs):
        result = ResolvedLink(current_view_name=None, link=self)
        result.separator = True
        return result


class SourceColumn(TemplateObjectMixin):
    _registry = {}

    @staticmethod
    def get_attribute_recursive(attribute, model):
        """
        Walk over the double underscore (__) separated path to the last
        field. Returns the field name and the corresponding model class.
        Used to introspect the label or short_description of a model's
        attribute.
        """
        last_model = model
        for part in attribute.split(LOOKUP_SEP):
            last_model = model
            try:
                field = model._meta.get_field(part)
            except FieldDoesNotExist:
                break
            else:
                model = field.related_model or field.model

        return part, last_model

    @staticmethod
    def sort(columns):
        columns.sort(key=lambda x: x.order)
        return columns

    @classmethod
    def get_column_matches(cls, source):
        columns = []

        try:
            # Try it as a queryset.
            model = source.model
        except AttributeError:
            try:
                # Try it as a list.
                item = source[0]
            except TypeError:
                # Neither a queryset nor a list.
                try:
                    # Try as a model instance or model.
                    model = source._meta.model
                except AttributeError:
                    # Not a model instance.

                    # Try as subclass instance, check the class hierarchy.
                    for super_class in source.__class__.__mro__[:-1]:
                        columns.extend(cls._registry.get(super_class, ()))

                    return columns
                else:
                    # Get model columns.
                    columns.extend(
                        cls._registry.get(model, ())
                    )

                    # Get proxy columns.
                    # Remove the columns explicitly excluded.
                    # Execute after the root model columns to allow a proxy
                    # to override an existing column.
                    for proxy_column in cls._registry.get(model._meta.proxy_for_model, ()):
                        if model not in proxy_column.excludes:
                            columns.append(proxy_column)

                    return columns
            else:
                # It was is a list.
                return cls.get_column_matches(source=item)
        else:
            # It was is a queryset.
            return cls.get_column_matches(source=model)

    @classmethod
    def get_for_source(
        cls, source, exclude_identifier=False, names=None,
        only_identifier=False
    ):
        # Process columns as a set to avoid duplicate resolved column
        # detection code.
        columns = cls.get_column_matches(source=source)

        if exclude_identifier:
            columns = [column for column in columns if not column.is_identifier]
        else:
            # exclude_identifier and only_identifier and mutually exclusive.
            if only_identifier:
                for column in columns:
                    if column.is_identifier:
                        return (column,)

                # There is no column with the identifier marker.
                return ()

        if names is not None:
            indexed_columns = {
                column.name: column for column in columns
            }

            return [indexed_columns[name] for name in names]

        columns = SourceColumn.sort(columns=columns)

        return columns

    def __init__(
        self, source, attribute=None, empty_value=None, func=None,
        help_text=None, html_extra_classes=None, include_label=False,
        is_attribute_absolute_url=False, is_object_absolute_url=False,
        is_identifier=False, is_sortable=False, kwargs=None, label=None,
        name=None, order=None, sort_field=None, widget=None
    ):
        """
        name: optional unique identifier for this source column for the
        specified source.
        """
        self._label = label
        self._help_text = help_text
        self.source = source
        self.attribute = attribute
        self.empty_value = empty_value
        self.excludes = set()
        self.func = func
        self.html_extra_classes = html_extra_classes or ''
        self.include_label = include_label
        self.is_attribute_absolute_url = is_attribute_absolute_url
        self.is_object_absolute_url = is_object_absolute_url
        self.is_identifier = is_identifier
        self.is_sortable = is_sortable
        self.kwargs = kwargs or {}
        self.name = name
        self.order = order or 0
        self.sort_field = sort_field
        self.widget = widget

        if self.is_attribute_absolute_url or self.is_object_absolute_url:
            if not self.widget:
                self.widget = SourceColumnLinkWidget

        self.__class__._registry.setdefault(source, [])
        self.__class__._registry[source].append(self)

        self._calculate_label()
        self._calculate_help_text()
        if self.is_sortable:
            field_name = self.sort_field or self.attribute
            try:
                get_related_field(
                    model=self.source, related_field_name=field_name
                )
            except FieldDoesNotExist as exception:
                raise ImproperlyConfigured(
                    '"{}" is not a field of "{}", cannot be used as a '
                    'sortable column.'.format(field_name, self.source)
                ) from exception

    def _calculate_help_text(self):
        if not self._help_text:
            if self.attribute:
                try:
                    attribute = resolve_attribute(
                        obj=self.source, attribute=self.attribute
                    )
                    self._help_text = getattr(attribute, 'help_text')
                except AttributeError:
                    try:
                        name, model = SourceColumn.get_attribute_recursive(
                            attribute=self.attribute, model=self.source._meta.model
                        )
                        self._help_text = help_text_for_field(
                            name=name, model=model
                        )
                    except AttributeError:
                        self._help_text = None

        self.help_text = self._help_text

    def _calculate_label(self):
        if not self._label:
            if self.attribute:
                try:
                    attribute = resolve_attribute(
                        obj=self.source, attribute=self.attribute
                    )
                    self._label = getattr(attribute, 'short_description')
                except AttributeError:
                    try:
                        name, model = SourceColumn.get_attribute_recursive(
                            attribute=self.attribute, model=self.source._meta.model
                        )
                        self._label = label_for_field(
                            name=name, model=model
                        )
                    except AttributeError:
                        self._label = self.attribute
            else:
                self._label = getattr(
                    self.func, 'short_description', _('Unnamed function')
                )

        self.label = self._label

    def add_exclude(self, source):
        self.excludes.add(source)

    def get_absolute_url(self, obj):
        if self.is_object_absolute_url:
            return obj.get_absolute_url()
        elif self.is_attribute_absolute_url:
            result = resolve_attribute(
                attribute=self.attribute, kwargs=self.kwargs,
                obj=obj
            )
            if result:
                return result.get_absolute_url()

    def get_previous_sort_fields(self, context):
        previous_sort_fields = context.get(TEXT_SORT_FIELD_VARIABLE_NAME, None)

        if previous_sort_fields:
            previous_sort_fields = previous_sort_fields.split(',')
        else:
            previous_sort_fields = ()

        return previous_sort_fields

    def get_sort_field(self):
        if self.sort_field:
            return self.sort_field
        else:
            return self.attribute

    def get_sort_field_querystring(self, context):
        request = self.get_request(context=context)

        # We do this to get an mutable copy we can modify.
        querystring = request.GET.copy()

        previous_sort_fields = self.get_previous_sort_fields(context=context)

        sort_field = self.get_sort_field()

        if sort_field in previous_sort_fields:
            result = '-{}'.format(sort_field)
        else:
            result = '{}'.format(sort_field)

        querystring[TEXT_SORT_FIELD_PARAMETER] = result

        return '?{}'.format(querystring.urlencode())

    def get_sort_icon(self, context):
        previous_sort_fields = self.get_previous_sort_fields(context=context)
        sort_field = self.get_sort_field()

        if sort_field in previous_sort_fields:
            return icon_sort_down
        elif '-{}'.format(sort_field) in previous_sort_fields:
            return icon_sort_up

    def resolve(self, context):
        if self.attribute:
            result = resolve_attribute(
                attribute=self.attribute, kwargs=self.kwargs,
                obj=context['object']
            )
        elif self.func:
            result = self.func(context=context, **self.kwargs)
        else:
            result = context['object']

        self.absolute_url = self.get_absolute_url(obj=context['object'])
        if self.widget:
            try:
                request = self.get_request(context=context)
            except VariableDoesNotExist:
                """
                Don't attempt to render and return the value if any.
                """
            else:
                widget_instance = self.widget(
                    column=self, request=request
                )
                return widget_instance.render(value=result)

        if not result:
            if self.empty_value:
                return self.empty_value
            else:
                return result
        else:
            return result


class Text(Link):
    """
    Menu text. Renders to a plain <li> tag.
    """
    def __init__(self, *args, **kwargs):
        self.html_extra_classes = kwargs.get('html_extra_classes', '')
        self.icon = None
        self.text = kwargs.get('text')
        self.view = None

    def resolve(self, *args, **kwargs):
        result = ResolvedLink(current_view_name=None, link=self)
        result.context = kwargs.get('context')
        result.text_span = True
        return result



import json
import logging

from django import forms
from django.db.models import Model
from django.db.models.query import QuerySet
from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.forms.document_forms import DocumentForm
from mayan.apps.documents.literals import DOCUMENT_FILE_ACTION_PAGE_CHOICES
from mayan.apps.views.forms import DynamicModelForm

from .classes import SourceBackend
from .models import Source

logger = logging.getLogger(name=__name__)


class NewDocumentForm(DocumentForm):
    class Meta(DocumentForm.Meta):
        exclude = ('label', 'description')


class NewDocumentFileForm(forms.Form):
    comment = forms.CharField(
        help_text=_('An optional comment to explain the upload.'),
        label=_('Comment'), required=False,
        widget=forms.widgets.Textarea(attrs={'rows': 4}),
    )
    action = forms.ChoiceField(
        choices=DOCUMENT_FILE_ACTION_PAGE_CHOICES, label=_('Action'),
        help_text=_(
            'The action to take in regards to the pages of the new file '
            'being uploaded.'
        )
    )


class UploadBaseForm(forms.Form):
    def __init__(self, *args, **kwargs):
        self.source = kwargs.pop('source')

        super().__init__(*args, **kwargs)


class SourceBackendSelectionForm(forms.Form):
    backend = forms.ChoiceField(
        choices=(), help_text=_('The backend used to create the new source.'),
        label=_('Backend')
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['backend'].choices = SourceBackend.get_choices()


class SourceBackendDynamicForm(DynamicModelForm):
    class Meta:
        fields = ('label', 'enabled', 'backend_data')
        model = Source
        widgets = {'backend_data': forms.widgets.HiddenInput}

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        if self.instance.backend_data:
            backend_data = json.loads(s=self.instance.backend_data)

            for key in self.instance.get_backend().get_fields():
                self.fields[key].initial = backend_data.get(key, None)

    def clean(self):
        data = super().clean()

        # Consolidate the dynamic fields into a single JSON field called
        # 'backend_data'.
        backend_data = {}

        for field_name, field_data in self.schema['fields'].items():
            backend_data[field_name] = data.pop(
                field_name, field_data.get('default', None)
            )
            if isinstance(backend_data[field_name], QuerySet):
                # Flatten the queryset to a list of ids
                backend_data[field_name] = list(
                    backend_data[field_name].values_list('id', flat=True)
                )
            elif isinstance(backend_data[field_name], Model):
                # Store only the ID of a model instance
                backend_data[field_name] = backend_data[field_name].pk

        data['backend_data'] = json.dumps(obj=backend_data)

        return data


class WebFormUploadFormHTML5(UploadBaseForm):
    file = forms.FileField(
        label=_('File'), widget=forms.widgets.FileInput(
            attrs={'class': 'hidden', 'hidden': True}
        )
    )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_AJAX_REDIRECTION_CODE, DEFAULT_MAXIMUM_TITLE_LENGTH,
    DEFAULT_MESSAGE_POSITION, DEFAULT_MENU_POLLING_INTERVAL
)

namespace = SettingNamespace(label=_('Appearance'), name='appearance')

setting_ajax_redirection_code = namespace.add_setting(
    default=DEFAULT_AJAX_REDIRECTION_CODE,
    global_name='APPEARANCE_AJAX_REDIRECTION_CODE', help_text=_(
        'Custom HTTP response code for AJAX redirections.'
    )
)
setting_max_title_length = namespace.add_setting(
    default=DEFAULT_MAXIMUM_TITLE_LENGTH,
    global_name='APPEARANCE_MAXIMUM_TITLE_LENGTH', help_text=_(
        'Maximum number of characters that will be displayed as the view '
        'title.'
    )
)
setting_message_position = namespace.add_setting(
    default=DEFAULT_MESSAGE_POSITION,
    global_name='APPEARANCE_MESSAGE_POSITION', help_text=_(
        'Position where the system message will be displayed. Options are: '
        'top-left, top-center, top-right, bottom-left, bottom-center, '
        'bottom-right.'
    )
)
setting_menu_polling_interval = namespace.add_setting(
    default=DEFAULT_MENU_POLLING_INTERVAL,
    global_name='APPEARANCE_MENU_POLLING_INTERVAL', help_text=_(
        'Delay in milliseconds after which the menus will be checked for '
        'updates.'
    )
)



from mayan.apps.documents.permissions import (
    permission_document_create, permission_document_file_new
)
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from ..links import link_document_create_multiple, link_document_file_upload

from .mixins.base_mixins import SourceTestMixin


class SourcesNewDocumentLinkTestCase(
    SourceTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def _get_document_create_link(self):
        self.add_test_view()
        context = self.get_test_view()
        context['user'] = self._test_case_user
        return link_document_create_multiple.resolve(context=context)

    def test_document_create_link_no_permission(self):
        resolved_link = self._get_document_create_link()
        self.assertEqual(resolved_link, None)

    def test_document_create_link_with_document_type_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_create
        )

        resolved_link = self._get_document_create_link()
        self.assertEqual(resolved_link, None)

    def test_document_create_link_with_source_access(self):
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        resolved_link = self._get_document_create_link()
        self.assertEqual(resolved_link, None)

    def test_document_create_link_with_full_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_create
        )

        resolved_link = self._get_document_create_link()
        self.assertNotEqual(resolved_link, None)


class SourcesNewDocumentFileLinkTestCase(
    SourceTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def _get_document_file_upload_link(self):
        self.add_test_view(test_object=self.test_document)
        context = self.get_test_view()
        context['user'] = self._test_case_user
        return link_document_file_upload.resolve(context=context)

    def test_document_file_upload_link_no_permission(self):
        resolved_link = self._get_document_file_upload_link()
        self.assertEqual(resolved_link, None)

    def test_document_file_upload_link_with_document_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_new
        )

        resolved_link = self._get_document_file_upload_link()
        self.assertEqual(resolved_link, None)

    def test_document_file_upload_link_with_source_access(self):
        self.grant_access(
            obj=self.test_source, permission=permission_document_file_new
        )

        resolved_link = self._get_document_file_upload_link()
        self.assertEqual(resolved_link, None)

    def test_document_file_upload_link_with_full_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_new
        )
        self.grant_access(
            obj=self.test_source, permission=permission_document_file_new
        )

        resolved_link = self._get_document_file_upload_link()
        self.assertNotEqual(resolved_link, None)



from django.contrib import messages
from django.http import HttpResponseRedirect
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _, ungettext

from mayan.apps.documents.forms.document_type_forms import (
    DocumentTypeFilteredSelectForm
)
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.models.document_file_page_models import DocumentFilePage
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.views.generics import (
    FormView, MultipleObjectConfirmActionView, MultipleObjectDeleteView,
    SingleObjectDetailView, SingleObjectDownloadView, SingleObjectEditView,
    SingleObjectListView
)
from mayan.apps.views.mixins import ExternalObjectViewMixin

from .forms import DocumentFileContentForm, DocumentFilePageContentForm
from .models import DocumentFileParseError, DocumentFilePageContent
from .permissions import (
    permission_document_file_content_view, permission_document_type_parsing_setup,
    permission_document_file_parse
)


class DocumentFileContentDeleteView(MultipleObjectDeleteView):
    error_message = _(
        'Error deleting document version content "%(instance)s"; %(exception)s'
    )
    object_permission = permission_document_file_parse
    pk_url_kwarg = 'document_file_id'
    source_queryset = DocumentFile.valid.all()
    success_message_single = _(
        'Content of "%(object)s" deleted successfully.'
    )
    success_message_singular = _(
        'Content of %(count)d document version deleted successfully.'
    )
    success_message_plural = _(
        'Content of %(count)d document versions deleted successfully.'
    )
    title_single = _('Delete the content of: %(object)s.')
    title_singular = _(
        'Delete the content of the %(count)d selected document version.'
    )
    title_plural = _(
        'Delete the content of the %(count)d selected document versions.'
    )

    def object_action(self, form, instance):
        DocumentFilePageContent.objects.delete_content_for(
            document_file=instance, user=self.request.user
        )


class DocumentFileContentDownloadView(SingleObjectDownloadView):
    object_permission = permission_document_file_content_view
    pk_url_kwarg = 'document_file_id'
    source_queryset = DocumentFile.valid.all()

    def get_download_file_object(self):
        return self.object.content()

    def get_download_filename(self):
        return '{}-content'.format(self.object)


class DocumentFileContentView(SingleObjectDetailView):
    form_class = DocumentFileContentForm
    object_permission = permission_document_file_content_view
    pk_url_kwarg = 'document_file_id'
    source_queryset = DocumentFile.valid.all()

    def dispatch(self, request, *args, **kwargs):
        result = super().dispatch(request=request, *args, **kwargs)
        self.object.document.add_as_recent_document_for_user(
            user=request.user
        )
        return result

    def get_extra_context(self):
        return {
            'hide_labels': True,
            'object': self.object,
            'title': _('Content for document file: %s') % self.object
        }


class DocumentFilePageContentView(SingleObjectDetailView):
    form_class = DocumentFilePageContentForm
    object_permission = permission_document_file_content_view
    pk_url_kwarg = 'document_file_page_id'
    source_queryset = DocumentFilePage.valid.all()

    def dispatch(self, request, *args, **kwargs):
        result = super().dispatch(request=request, *args, **kwargs)
        self.object.document_file.document.add_as_recent_document_for_user(
            request.user
        )
        return result

    def get_extra_context(self):
        return {
            'hide_labels': True,
            'object': self.object,
            'title': _('Content for document file page: %s') % self.object,
        }

    def get_source_queryset(self):
        document_file_queryset = DocumentFile.valid.all()
        return DocumentFilePage.objects.filter(
            document_file_id__in=document_file_queryset.values('pk')
        )


class DocumentFileParsingErrorsListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_permission = permission_document_file_parse
    external_object_pk_url_kwarg = 'document_file_id'
    external_object_queryset = DocumentFile.valid.all()

    def get_extra_context(self):
        return {
            'hide_object': True,
            'object': self.external_object,
            'title': _(
                'Parsing errors for document file: %s'
            ) % self.external_object,
        }

    def get_source_queryset(self):
        return self.external_object.parsing_errors.all()


class DocumentFileSubmitView(MultipleObjectConfirmActionView):
    object_permission = permission_document_file_parse
    pk_url_kwarg = 'document_file_id'
    source_queryset = DocumentFile.valid.all()
    success_message = _(
        '%(count)d document file added to the parsing queue'
    )
    success_message_plural = _(
        '%(count)d documents files added to the parsing queue'
    )

    def get_extra_context(self):
        queryset = self.object_list

        result = {
            'title': ungettext(
                singular='Submit %(count)d document file to the parsing queue?',
                plural='Submit %(count)d documents files to the parsing queue?',
                number=queryset.count()
            ) % {
                'count': queryset.count(),
            }
        }

        if queryset.count() == 1:
            result.update(
                {
                    'object': queryset.first(),
                    'title': _(
                        'Submit document file "%s" to the parsing queue'
                    ) % queryset.first()
                }
            )

        return result

    def object_action(self, instance, form=None):
        instance.submit_for_parsing(_user=self.request.user)


class DocumentTypeSettingsEditView(ExternalObjectViewMixin, SingleObjectEditView):
    external_object_class = DocumentType
    external_object_permission = permission_document_type_parsing_setup
    external_object_pk_url_kwarg = 'document_type_id'
    fields = ('auto_parsing',)
    post_action_redirect = reverse_lazy(
        viewname='documents:document_type_list'
    )

    def get_document_type(self):
        return self.external_object

    def get_extra_context(self):
        return {
            'object': self.get_document_type(),
            'title': _(
                'Edit parsing settings for document type: %s.'
            ) % self.get_document_type()
        }

    def get_object(self, queryset=None):
        return self.get_document_type().parsing_settings


class DocumentTypeSubmitView(FormView):
    extra_context = {
        'title': _('Submit all documents of a type for parsing')
    }
    form_class = DocumentTypeFilteredSelectForm
    post_action_redirect = reverse_lazy(viewname='common:tools_list')

    def get_form_extra_kwargs(self):
        return {
            'allow_multiple': True,
            'permission': permission_document_file_parse,
            'user': self.request.user
        }

    def form_valid(self, form):
        count = 0
        for document_type in form.cleaned_data['document_type']:
            for document in document_type.documents.all():
                document.submit_for_parsing(_user=self.request.user)
                count += 1

        messages.success(
            message=_(
                '%(count)d documents added to the parsing queue.'
            ) % {
                'count': count,
            }, request=self.request
        )

        return HttpResponseRedirect(redirect_to=self.get_success_url())


class ParseErrorListView(SingleObjectListView):
    extra_context = {
        'hide_object': True,
        'title': _('Parsing errors'),
    }
    view_permission = permission_document_file_parse

    def get_source_queryset(self):
        return DocumentFileParseError.objects.all()



import logging

from django.urls import resolve as django_resolve
from django.urls.base import get_script_prefix
from django.utils.encoding import force_text

logger = logging.getLogger(name=__name__)


def convert_to_id_list(items):
    return ','.join(map(force_text, items))


def resolve(path, urlconf=None):
    path = '/{}'.format(path.replace(get_script_prefix(), '', 1))
    return django_resolve(path=path, urlconf=urlconf)



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sender_object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('subject', models.CharField(help_text='Short description of this message.', max_length=255, verbose_name='Subject')),
                ('body', models.TextField(help_text='The actual content of the message.', verbose_name='Body')),
                ('read', models.BooleanField(default=False, help_text='This field determines if the message has been read or not.', verbose_name='Read')),
                ('date_time', models.DateTimeField(auto_now_add=True, help_text='Date and time of the message creation.', verbose_name='Creation date and time')),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='messages', to='messaging.Message', verbose_name='Parent message')),
                ('sender_content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='messages', to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Message',
                'verbose_name_plural': 'Messages',
                'ordering': ('-date_time',),
            },
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('linking', '0006_auto_20180402_0339'),
    ]

    operations = [
        migrations.AlterField(
            model_name='smartlink',
            name='dynamic_label',
            field=models.CharField(
                blank=True, help_text="Enter a template to render. Use "
                "Django's default templating language "
                "(https://docs.djangoproject.com/en/1.11/ref/templates/builtins/). "
                "The {{ document }} context variable is available.",
                max_length=96, verbose_name='Dynamic label'
            ),
        ),
        migrations.AlterField(
            model_name='smartlinkcondition',
            name='expression',
            field=models.TextField(
                help_text="Enter a template to render. Use Django's default "
                "templating language "
                "(https://docs.djangoproject.com/en/1.11/ref/templates/builtins/). "
                "The {{ document }} context variable is available.",
                verbose_name='Expression'
            ),
        ),
    ]



from django.apps import apps
from django.db.models.signals import post_migrate
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import permission_acl_edit, permission_acl_view
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import (
    MissingItem, ModelCopy, ModelField, ModelFieldRelated, ModelProperty,
    ModelQueryFields
)
from mayan.apps.common.menus import (
    menu_facet, menu_list_facet, menu_main, menu_object, menu_return,
    menu_secondary, menu_setup, menu_multi_item
)
from mayan.apps.common.signals import signal_post_initial_setup
from mayan.apps.converter.classes import AppImageErrorImage
from mayan.apps.converter.layers import layer_decorations
from mayan.apps.converter.links import link_transformation_list
from mayan.apps.converter.permissions import (
    permission_transformation_create,
    permission_transformation_delete, permission_transformation_edit,
    permission_transformation_view,
)
from mayan.apps.dashboards.dashboards import dashboard_administrator
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.file_caching.links import link_cache_partition_purge
from mayan.apps.file_caching.permissions import permission_cache_partition_purge
from mayan.apps.navigation.classes import SourceColumn
from mayan.apps.rest_api.fields import DynamicSerializerField
from mayan.apps.templating.classes import AJAXTemplate
from mayan.apps.user_management.dashboards import dashboard_user
from mayan.apps.views.html_widgets import TwoStateWidget

from .dashboard_widgets import (
    DashboardWidgetDocumentFilePagesTotal, DashboardWidgetDocumentsInTrash,
    DashboardWidgetDocumentsNewThisMonth,
    DashboardWidgetDocumentsPagesNewThisMonth, DashboardWidgetDocumentsTotal,
    DashboardWidgetDocumentsTypesTotal,
    DashboardWidgetUserRecentlyAccessedDocuments,
    DashboardWidgetUserRecentlyCreatedDocuments,
    DashboardWidgetUserFavoriteDocuments
)

# Documents

from .events import (
    event_document_created, event_document_edited, event_document_trashed,
    event_document_viewed, event_trashed_document_deleted,
    event_trashed_document_restored
)

# Document files

from .events import (
    event_document_file_created, event_document_file_deleted,
    event_document_file_downloaded, event_document_file_edited
)

# Document types

from .events import (
    event_document_type_changed, event_document_type_edited,
    event_document_type_quick_label_created,
    event_document_type_quick_label_deleted,
    event_document_type_quick_label_edited
)

# Document versions

from .events import (
    event_document_version_created, event_document_version_deleted,
    event_document_version_edited, event_document_version_exported,
    event_document_version_page_created, event_document_version_page_deleted,
    event_document_version_page_edited,
)

from .handlers import (
    handler_create_default_document_type,
    handler_create_document_file_page_image_cache,
    handler_create_document_version_page_image_cache
)
from .html_widgets import ThumbnailWidget
from .links.document_links import (
    link_document_type_change, link_document_properties_edit,
    link_document_list, link_document_recently_accessed_list,
    link_document_recently_created_list, link_document_multiple_type_change,
    link_document_preview, link_document_properties
)
from .links.document_file_links import (
    link_document_file_delete, link_document_file_delete_multiple,
    link_document_file_download_quick, link_document_file_edit,
    link_document_file_list, link_document_file_preview,
    link_document_file_print_form, link_document_file_properties,
    link_document_file_return_to_document, link_document_file_return_list,
    link_document_file_transformations_clear,
    link_document_file_multiple_transformations_clear,
    link_document_file_transformations_clone
)
from .links.document_file_page_links import (
    link_document_file_multiple_page_count_update,
    link_document_file_page_count_update, link_document_file_page_list,
    link_document_file_page_navigation_first,
    link_document_file_page_navigation_last,
    link_document_file_page_navigation_next,
    link_document_file_page_navigation_previous,
    link_document_file_page_return_to_document,
    link_document_file_page_return_to_document_file,
    link_document_file_page_return_to_document_file_page_list,
    link_document_file_page_rotate_left, link_document_file_page_rotate_right,
    link_document_file_page_view, link_document_file_page_view_reset,
    link_document_file_page_zoom_in, link_document_file_page_zoom_out
)
from .links.document_type_links import (
    link_document_type_create, link_document_type_delete,
    link_document_type_edit, link_document_type_filename_create,
    link_document_type_filename_delete, link_document_type_filename_edit,
    link_document_type_filename_list, link_document_type_filename_generator,
    link_document_type_list, link_document_type_policies,
    link_document_type_setup
)
from .links.document_version_links import (
    link_document_version_active, link_document_version_create,
    link_document_version_delete_single,
    link_document_version_delete_multiple, link_document_version_edit,
    link_document_version_export, link_document_version_list,
    link_document_version_return_list,
    link_document_version_return_to_document, link_document_version_preview,
    link_document_version_print_form,
    link_document_version_transformations_clear,
    link_document_version_multiple_transformations_clear,
    link_document_version_transformations_clone
)
from .links.document_version_page_links import (
    link_document_version_page_delete, link_document_version_page_list,
    link_document_version_page_list_append,
    link_document_version_page_list_remap,
    link_document_version_page_list_reset,
    link_document_version_page_navigation_first,
    link_document_version_page_navigation_last,
    link_document_version_page_navigation_next,
    link_document_version_page_navigation_previous,
    link_document_version_page_return_to_document,
    link_document_version_page_return_to_document_version,
    link_document_version_page_return_to_document_version_page_list,
    link_document_version_page_rotate_left,
    link_document_version_page_rotate_right, link_document_version_page_view,
    link_document_version_page_view_reset,
    link_document_version_page_zoom_in, link_document_version_page_zoom_out
)
from .links.favorite_links import (
    link_document_favorites_add, link_document_favorites_remove,
    link_document_favorites_list, link_document_favorites_add_multiple,
    link_document_favorites_remove_multiple
)
from .links.trashed_document_links import (
    link_document_delete, link_document_list_deleted,
    link_document_multiple_delete, link_document_multiple_restore,
    link_document_multiple_trash, link_document_restore, link_document_trash,
    link_trash_can_empty
)
from .literals import (
    IMAGE_ERROR_NO_ACTIVE_VERSION, IMAGE_ERROR_NO_VERSION_PAGES,
    IMAGE_ERROR_FILE_PAGE_TRANSFORMATION_ERROR,
    IMAGE_ERROR_VERSION_PAGE_TRANSFORMATION_ERROR
)
from .menus import menu_documents

# Documents

from .permissions import (
    permission_document_create, permission_document_edit,
    permission_document_properties_edit, permission_document_tools,
    permission_document_trash, permission_document_view
)

# DocumentFile

from .permissions import (
    permission_document_file_delete, permission_document_file_download,
    permission_document_file_edit, permission_document_file_new,
    permission_document_file_print, permission_document_file_tools,
    permission_document_file_view
)

# DocumentType

from .permissions import (
    permission_document_type_delete, permission_document_type_edit,
    permission_document_type_view
)

# DocumentVersion

from .permissions import (
    permission_document_version_create, permission_document_version_delete,
    permission_document_version_edit, permission_document_version_export,
    permission_document_version_print, permission_document_version_view
)

# TrashedDocument

from .permissions import (
    permission_trashed_document_delete, permission_trashed_document_restore
)

from .statistics import *  # NOQA


class DocumentsApp(MayanAppConfig):
    app_namespace = 'documents'
    app_url = 'documents'
    has_rest_api = True
    has_static_media = True
    has_tests = True
    name = 'mayan.apps.documents'
    verbose_name = _('Documents')

    def ready(self):
        super().ready()

        Document = self.get_model(model_name='Document')
        DocumentFile = self.get_model(model_name='DocumentFile')
        DocumentFilePage = self.get_model(model_name='DocumentFilePage')
        DocumentType = self.get_model(model_name='DocumentType')
        DocumentTypeFilename = self.get_model(
            model_name='DocumentTypeFilename'
        )
        DocumentVersion = self.get_model(model_name='DocumentVersion')
        DocumentVersionPage = self.get_model(model_name='DocumentVersionPage')
        DocumentVersionPageSearchResult = self.get_model(
            model_name='DocumentVersionPageSearchResult'
        )
        DownloadFile = apps.get_model(
            app_label='storage', model_name='DownloadFile'
        )
        FavoriteDocument = self.get_model(
            model_name='FavoriteDocument'
        )
        FavoriteDocumentProxy = self.get_model(
            model_name='FavoriteDocumentProxy'
        )
        RecentlyAccessedDocument = self.get_model(
            model_name='RecentlyAccessedDocument'
        )
        RecentlyAccessedDocumentProxy = self.get_model(
            model_name='RecentlyAccessedDocumentProxy'
        )
        RecentlyCreatedDocument = self.get_model(
            model_name='RecentlyCreatedDocument'
        )
        TrashedDocument = self.get_model(model_name='TrashedDocument')

        AppImageErrorImage(
            name=IMAGE_ERROR_NO_ACTIVE_VERSION,
            template_name='documents/errors/no_valid_version.html'
        )
        AppImageErrorImage(
            name=IMAGE_ERROR_NO_VERSION_PAGES,
            template_name='documents/errors/no_version_pages.html'
        )
        AppImageErrorImage(
            name=IMAGE_ERROR_FILE_PAGE_TRANSFORMATION_ERROR,
            template_name='documents/errors/document_file_page_transformation_error.html'
        )
        AppImageErrorImage(
            name=IMAGE_ERROR_VERSION_PAGE_TRANSFORMATION_ERROR,
            template_name='documents/errors/document_version_page_transformation_error.html'
        )

        AJAXTemplate(
            name='invalid_document',
            template_name='documents/invalid_document.html'
        )

        link_decorations_list = link_transformation_list.copy(
            layer=layer_decorations
        )
        link_decorations_list.text = _('Decorations')

        DownloadFile.objects.register_content_object(model=DocumentVersion)

        DynamicSerializerField.add_serializer(
            klass=Document,
            serializer_class='mayan.apps.documents.serializers.document_serializers.DocumentSerializer'
        )

        EventModelRegistry.register(model=Document)
        EventModelRegistry.register(model=DocumentFile)
        EventModelRegistry.register(model=DocumentFilePage)
        EventModelRegistry.register(model=DocumentType)
        EventModelRegistry.register(model=DocumentTypeFilename)
        EventModelRegistry.register(model=DocumentVersion)
        EventModelRegistry.register(model=DocumentVersionPage)
        EventModelRegistry.register(model=TrashedDocument)

        MissingItem(
            label=_('Create a document type'),
            description=_(
                'Every uploaded document must be assigned a document type, '
                'it is the basic way Mayan EDMS categorizes documents.'
            ), condition=lambda: not DocumentType.objects.exists(),
            view='documents:document_type_list'
        )

        ModelCopy(model=DocumentTypeFilename).add_fields(
            field_names=(
                'document_type', 'filename', 'enabled'
            )
        )
        ModelCopy(
            model=DocumentType, bind_link=True, register_permission=True
        ).add_fields(
            field_names=(
                'label', 'trash_time_period', 'trash_time_unit',
                'delete_time_period', 'delete_time_unit', 'filenames'
            )
        )
        ModelCopy(
            model=DocumentVersion, bind_link=True, register_permission=True
        ).add_fields(
            field_names=(
                'document', 'timestamp', 'comment', 'version_pages',
            )
        )
        ModelCopy(
            model=DocumentVersionPage, bind_link=True, register_permission=True
        ).add_fields(
            field_names=(
                'document_version', 'page_number', 'content_type', 'object_id',
            )
        )

        ModelEventType.register(
            model=Document, event_types=(
                event_document_edited,
                event_document_type_changed,
                event_document_file_created,
                event_document_file_edited,
                event_document_file_deleted,
                event_document_version_created,
                event_document_version_edited,
                event_document_version_deleted,
                event_document_viewed,
                event_document_trashed,
                event_trashed_document_restored
            )
        )
        ModelEventType.register(
            model=DocumentFile, event_types=(
                event_document_file_downloaded,
                event_document_file_edited,
            )
        )
        ModelEventType.register(
            model=DocumentType, event_types=(
                event_document_created,
                event_document_type_edited,
                event_document_type_quick_label_created,
                event_trashed_document_deleted
            )
        )
        ModelEventType.register(
            model=DocumentTypeFilename, event_types=(
                event_document_type_quick_label_deleted,
                event_document_type_quick_label_edited
            )
        )
        ModelEventType.register(
            model=DocumentVersion, event_types=(
                event_document_version_edited,
                event_document_version_exported,
                event_document_version_page_created,
                event_document_version_page_deleted
            )
        )
        ModelEventType.register(
            model=DocumentVersionPage, event_types=(
                event_document_version_page_edited,
            )
        )

        ModelField(model=Document, name='description')
        ModelField(model=Document, name='datetime_created')
        ModelField(model=Document, name='trashed_date_time')
        ModelField(
            model=Document, name='document_type'
        )
        ModelField(model=Document, name='in_trash')
        ModelField(model=Document, name='is_stub')
        ModelField(model=Document, name='label')
        ModelField(model=Document, name='language')
        ModelField(model=Document, name='uuid')

        ModelFieldRelated(model=Document, name='document_type__label')
        ModelFieldRelated(
            model=Document,
            name='files__checksum'
        )
        ModelFieldRelated(
            model=Document, label=_('File comments'),
            name='files__comment'
        )
        ModelFieldRelated(
            model=Document, label=_('File encodings'),
            name='files__encoding'
        )
        ModelFieldRelated(
            model=Document, label=_('File MIME types'),
            name='files__mimetype'
        )
        ModelFieldRelated(
            model=Document, label=_('File timestamps'),
            name='files__timestamp'
        )

        ModelField(
            model=DocumentFilePage, label=_('Document file'),
            name='document_file'
        )
        ModelField(
            model=DocumentFilePage, label=_('Page number'),
            name='page_number'
        )

        ModelProperty(
            description=_('Return the latest file of the document.'),
            model=Document, label=_('Latest file'), name='latest_file'
        )
        ModelProperty(
            description=_('Return the document instance.'),
            model=DocumentFilePage, label=_('Document'), name='document'
        )

        ModelPermission.register(
            model=Document, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_document_edit, permission_document_file_new,
                permission_document_properties_edit,
                permission_document_tools,
                permission_document_trash, permission_document_view,
                permission_document_version_create,
                permission_trashed_document_delete,
                permission_trashed_document_restore
            )
        )
        ModelPermission.register(
            model=DocumentFile, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_cache_partition_purge,
                permission_document_file_delete,
                permission_document_file_download,
                permission_document_file_edit,
                permission_document_file_print,
                permission_document_file_tools,
                permission_document_file_view,
                permission_transformation_create,
                permission_transformation_delete,
                permission_transformation_edit,
                permission_transformation_view
            )
        )
        ModelPermission.register(
            model=DocumentType, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_document_create, permission_document_type_delete,
                permission_document_type_edit, permission_document_type_view
            )
        )
        ModelPermission.register(
            model=DocumentVersion, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_cache_partition_purge,
                permission_document_version_delete,
                permission_document_version_edit,
                permission_document_version_export,
                permission_document_version_print,
                permission_document_version_view,
                permission_transformation_create,
                permission_transformation_delete,
                permission_transformation_edit,
                permission_transformation_view
            )
        )

        ModelPermission.register_inheritance(
            model=Document, related='document_type',
        )
        ModelPermission.register_inheritance(
            model=DocumentFile, related='document',
        )
        ModelPermission.register_inheritance(
            model=DocumentFile, related='document__document_type',
        )
        ModelPermission.register_inheritance(
            model=DocumentFilePage, related='document_file',
        )
        ModelPermission.register_inheritance(
            model=DocumentVersion, related='document',
        )
        ModelPermission.register_inheritance(
            model=DocumentVersion, related='document__document_type',
        )
        ModelPermission.register_inheritance(
            model=RecentlyAccessedDocument, related='document',
        )
        ModelPermission.register_inheritance(
            model=DocumentVersionPage, related='document_version',
        )
        ModelPermission.register_inheritance(
            model=DocumentTypeFilename, related='document_type',
        )
        ModelPermission.register_inheritance(
            model=FavoriteDocument, related='document',
        )

        model_query_fields_document = ModelQueryFields(model=Document)
        model_query_fields_document.add_prefetch_related_field(
            field_name='files'
        )
        model_query_fields_document.add_prefetch_related_field(
            field_name='files__file_pages'
        )
        model_query_fields_document.add_select_related_field(
            field_name='document_type'
        )

        model_query_fields_document_file = ModelQueryFields(model=DocumentFile)
        model_query_fields_document_file.add_prefetch_related_field(
            field_name='file_pages'
        )
        model_query_fields_document_file.add_select_related_field(
            field_name='document'
        )

        model_query_fields_document_file_page = ModelQueryFields(
            model=DocumentFilePage
        )
        model_query_fields_document_file_page.add_select_related_field(
            field_name='document_file'
        )

        model_query_fields_document_version = ModelQueryFields(
            model=DocumentVersion
        )
        model_query_fields_document_version.add_prefetch_related_field(
            field_name='version_pages'
        )
        model_query_fields_document_version.add_select_related_field(
            field_name='document'
        )

        # Document

        SourceColumn(
            attribute='get_label', is_object_absolute_url=True,
            is_identifier=True, is_sortable=True, name='label',
            sort_field='label', source=Document
        )
        SourceColumn(
            html_extra_classes='text-center document-thumbnail-list',
            label=_('Thumbnail'), order=-99, source=Document,
            widget=ThumbnailWidget
        )

        SourceColumn(
            attribute='document_type', include_label=True, is_sortable=True,
            label=_('Type'), name='document_type', order=-9, source=Document
        )
        SourceColumn(
            func=lambda context: context['object'].pages.count(),
            label=_('Pages'), include_label=True, order=-8, source=Document
        )

        # FavoriteDocumentProxy

        SourceColumn(
            func=lambda context: context['object'].favorites.get(
                user=context['request'].user
            ).datetime_added, include_label=True, is_sortable=True,
            label=_('Date and time added'), name='datetime_added',
            sort_field='favorites__datetime_added',
            source=FavoriteDocumentProxy
        )

        # RecentlyAccessedDocument

        SourceColumn(
            func=lambda context: context['object'].recent.first().datetime_accessed,
            include_label=True,
            is_sortable=True,
            label=_('Access date and time'),
            name='datetime_accessed',
            sort_field='recent__datetime_accessed',
            source=RecentlyAccessedDocumentProxy
        )

        # RecentlyCreatedDocument

        SourceColumn(
            attribute='datetime_created', include_label=True,
            is_sortable=True, name='datetime_created',
            source=RecentlyCreatedDocument
        )

        # DocumentFile

        SourceColumn(
            source=DocumentFile, attribute='filename', is_identifier=True,
            is_object_absolute_url=True
        )
        SourceColumn(
            html_extra_classes='text-center document-thumbnail-list',
            label=_('Thumbnail'), order=-99, source=DocumentFile,
            widget=ThumbnailWidget
        )
        SourceColumn(
            func=lambda context: context['object'].pages.count(),
            include_label=True, label=_('Pages'), order=-6,
            source=DocumentFile
        )
        SourceColumn(
            attribute='comment', is_sortable=True, order=-7,
            source=DocumentFile
        )
        SourceColumn(
            attribute='encoding', include_label=True, is_sortable=True,
            order=-8, source=DocumentFile
        )
        SourceColumn(
            attribute='mimetype', include_label=True, is_sortable=True,
            order=-9, source=DocumentFile
        )

        # DocumentFilePage

        SourceColumn(
            attribute='get_label', is_identifier=True,
            is_object_absolute_url=True, source=DocumentFilePage,
        )
        SourceColumn(
            html_extra_classes='text-center document-thumbnail-list',
            label=_('Thumbnail'), order=-99, source=DocumentFilePage,
            widget=ThumbnailWidget
        )

        # DocumentType

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=DocumentType
        )

        SourceColumn(
            func=lambda context: context['object'].get_document_count(
                user=context['request'].user
            ), include_label=True, label=_('Documents'), source=DocumentType
        )

        SourceColumn(
            attribute='filename', is_identifier=True, is_sortable=True,
            source=DocumentTypeFilename
        )
        SourceColumn(
            attribute='enabled', include_label=True, is_sortable=True,
            source=DocumentTypeFilename, widget=TwoStateWidget
        )

        # DocumentVersion

        SourceColumn(
            source=DocumentVersion, attribute='get_label', is_identifier=True,
            is_object_absolute_url=True
        )
        SourceColumn(
            html_extra_classes='text-center document-thumbnail-list',
            label=_('Thumbnail'), order=-99, source=DocumentVersion,
            widget=ThumbnailWidget
        )
        SourceColumn(
            func=lambda context: context['object'].pages.count(),
            include_label=True, label=_('Pages'), order=-8,
            source=DocumentVersion
        )
        SourceColumn(
            attribute='active', include_label=True, is_sortable=True,
            order=-9, source=DocumentVersion, widget=TwoStateWidget
        )
        SourceColumn(
            attribute='comment', include_label=True, is_sortable=True,
            order=-7, source=DocumentVersion
        )

        # DocumentVersionPage

        SourceColumn(
            attribute='get_label', is_identifier=True,
            is_object_absolute_url=True, source=DocumentVersionPage,
        )
        SourceColumn(
            html_extra_classes='text-center document-thumbnail-list',
            label=_('Thumbnail'), order=-99, source=DocumentVersionPage,
            widget=ThumbnailWidget
        )

        # TrashedDocument

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=TrashedDocument
        )
        SourceColumn(
            attribute='trashed_date_time', include_label=True, order=99,
            source=TrashedDocument
        )

        # Dashboards

        dashboard_administrator.add_widget(
            widget=DashboardWidgetDocumentsTotal, order=0
        )
        dashboard_administrator.add_widget(
            widget=DashboardWidgetDocumentFilePagesTotal, order=1
        )
        dashboard_administrator.add_widget(
            widget=DashboardWidgetDocumentsInTrash, order=2
        )
        dashboard_administrator.add_widget(
            widget=DashboardWidgetDocumentsTypesTotal, order=3
        )
        dashboard_administrator.add_widget(
            widget=DashboardWidgetDocumentsNewThisMonth, order=4
        )
        dashboard_administrator.add_widget(
            widget=DashboardWidgetDocumentsPagesNewThisMonth, order=5
        )

        dashboard_user.add_widget(
            widget=DashboardWidgetUserRecentlyAccessedDocuments, order=1
        )
        dashboard_user.add_widget(
            widget=DashboardWidgetUserRecentlyCreatedDocuments, order=2
        )
        dashboard_user.add_widget(
            widget=DashboardWidgetUserFavoriteDocuments, order=3
        )

        menu_documents.bind_links(
            links=(
                link_document_recently_accessed_list,
                link_document_recently_created_list, link_document_favorites_list,
                link_document_list, link_document_list_deleted
            )
        )

        menu_main.bind_links(links=(menu_documents,), position=10)

        menu_setup.bind_links(links=(link_document_type_setup,))

        # Document

        menu_list_facet.bind_links(
            links=(link_document_preview,), sources=(Document,), position=0
        )
        menu_list_facet.bind_links(
            links=(link_document_properties,), sources=(Document,), position=2
        )
        menu_list_facet.bind_links(
            links=(
                link_document_file_list, link_document_version_list
            ), sources=(Document,), position=2
        )

        menu_object.bind_links(
            links=(
                link_document_favorites_add, link_document_favorites_remove,
                link_document_properties_edit, link_document_type_change,
                link_document_trash
            ), sources=(Document,)
        )

        menu_multi_item.bind_links(
            links=(
                link_document_favorites_add_multiple,
                link_document_favorites_remove_multiple,
                link_document_multiple_trash,
                link_document_multiple_type_change
            ), sources=(Document,)
        )

        menu_secondary.bind_links(
            links=(link_document_version_create,),
            sources=(
                'documents:document_version_create',
                'documents:document_version_list'
            )
        )

        # DocumentFile

        menu_list_facet.bind_links(
            links=(
                link_document_file_page_list, link_document_file_properties,
                link_document_file_preview
            ), sources=(DocumentFile,)
        )
        menu_multi_item.bind_links(
            links=(
                link_document_file_delete_multiple,
                link_document_file_multiple_page_count_update,
                link_document_file_multiple_transformations_clear,
            ), sources=(DocumentFile,)
        )
        menu_object.bind_links(
            links=(
                link_cache_partition_purge,
                link_document_file_delete,
                link_document_file_download_quick,
                link_document_file_edit,
                link_document_file_page_count_update,
                link_document_file_print_form,
                link_document_file_transformations_clear,
                link_document_file_transformations_clone
            ),
            sources=(DocumentFile,)
        )
        menu_return.bind_links(
            links=(
                link_document_file_return_list,
                link_document_file_return_to_document,
            ), sources=(DocumentFile,)
        )

        # DocumentFilePages

        menu_facet.add_unsorted_source(source=DocumentFilePage)
        menu_facet.bind_links(
            links=(
                link_document_file_page_rotate_left,
                link_document_file_page_rotate_right, link_document_file_page_zoom_in,
                link_document_file_page_zoom_out, link_document_file_page_view_reset
            ), sources=('documents:document_file_page_view',)
        )
        menu_facet.bind_links(
            links=(
                link_document_file_page_view,
                link_document_file_page_navigation_first,
                link_document_file_page_navigation_previous,
                link_document_file_page_navigation_next,
                link_document_file_page_navigation_last
            ), sources=(DocumentFilePage,)
        )
        menu_list_facet.bind_links(
            links=(link_decorations_list, link_transformation_list),
            sources=(DocumentFilePage,)
        )
        menu_return.bind_links(
            links=(
                link_document_file_page_return_to_document,
                link_document_file_page_return_to_document_file,
                link_document_file_page_return_to_document_file_page_list
            ),
            sources=(DocumentFilePage,)
        )

        # DocumentType

        menu_list_facet.bind_links(
            links=(
                link_document_type_filename_list,
                link_document_type_policies,
                link_document_type_filename_generator
            ), sources=(DocumentType,)
        )

        menu_object.bind_links(
            links=(
                link_document_type_delete, link_document_type_edit
            ), sources=(DocumentType,)
        )
        menu_object.bind_links(
            links=(
                link_document_type_filename_edit,
                link_document_type_filename_delete
            ), sources=(DocumentTypeFilename,)
        )
        menu_secondary.bind_links(
            links=(link_document_type_list, link_document_type_create),
            sources=(
                DocumentType, 'documents:document_type_create',
                'documents:document_type_list'
            )
        )
        menu_secondary.bind_links(
            links=(link_document_type_filename_create,),
            sources=(
                DocumentTypeFilename, 'documents:document_type_filename_list',
                'documents:document_type_filename_create'
            )
        )
        menu_secondary.bind_links(
            links=(link_trash_can_empty,),
            sources=(
                'documents:document_list_deleted', 'documents:trash_can_empty'
            )
        )

        # DocumentVersion

        menu_list_facet.bind_links(
            links=(
                link_document_version_page_list,
                link_document_version_preview
            ),
            sources=(DocumentVersion,)
        )
        menu_multi_item.bind_links(
            links=(
                link_document_version_delete_multiple,
                link_document_version_multiple_transformations_clear,
            ), sources=(DocumentVersion,)
        )
        menu_object.bind_links(
            links=(
                link_document_version_active,
                link_cache_partition_purge,
                link_document_version_delete_single, link_document_version_edit,
                link_document_version_export,
                link_document_version_page_list_append,
                link_document_version_page_list_remap,
                link_document_version_page_list_reset,
                link_document_version_print_form,
                link_document_version_transformations_clear,
                link_document_version_transformations_clone
            ),
            sources=(DocumentVersion,)
        )
        menu_return.bind_links(
            links=(
                link_document_version_return_list,
                link_document_version_return_to_document,
            ), sources=(DocumentVersion,)
        )

        # DocumentVersionPage

        menu_facet.add_unsorted_source(source=DocumentVersionPage)
        menu_facet.bind_links(
            links=(
                link_document_version_page_rotate_left,
                link_document_version_page_rotate_right, link_document_version_page_zoom_in,
                link_document_version_page_zoom_out, link_document_version_page_view_reset
            ), sources=('documents:document_version_page_view',)
        )
        menu_facet.bind_links(
            links=(link_document_version_page_view,),
            sources=(DocumentVersionPage,)
        )
        menu_facet.bind_links(
            links=(
                link_document_version_page_navigation_first,
                link_document_version_page_navigation_previous,
                link_document_version_page_navigation_next,
                link_document_version_page_navigation_last
            ), sources=(DocumentVersionPage,)
        )
        menu_list_facet.bind_links(
            links=(
                link_decorations_list, link_transformation_list,
            ), sources=(DocumentVersionPage, DocumentVersionPageSearchResult)
        )
        menu_object.bind_links(
            links=(
                link_document_version_page_delete,
            ), sources=(DocumentVersionPage, DocumentVersionPageSearchResult)
        )
        menu_return.bind_links(
            links=(
                link_document_version_page_return_to_document,
                link_document_version_page_return_to_document_version,
                link_document_version_page_return_to_document_version_page_list
            ), sources=(DocumentVersionPage,)
        )

        # Trashed documents

        menu_object.bind_links(
            links=(link_document_restore, link_document_delete),
            sources=(TrashedDocument,)
        )
        menu_multi_item.bind_links(
            links=(
                link_document_multiple_restore, link_document_multiple_delete
            ), sources=(TrashedDocument,)
        )
        menu_multi_item.add_proxy_exclusion(source=TrashedDocument)

        post_migrate.connect(
            dispatch_uid='documents_handler_create_document_file_page_image_cache',
            receiver=handler_create_document_file_page_image_cache,
        )
        post_migrate.connect(
            dispatch_uid='documents_handler_create_document_version_page_image_cache',
            receiver=handler_create_document_version_page_image_cache,
        )
        signal_post_initial_setup.connect(
            dispatch_uid='documents_handler_create_default_document_type',
            receiver=handler_create_default_document_type
        )



import logging

from django.apps import apps
from django.db.models.aggregates import Max, Min

from mayan.apps.lock_manager.exceptions import LockError
from mayan.celery import app

from .classes import SearchBackend, SearchModel
from .exceptions import DynamicSearchException, DynamicSearchRetry
from .literals import (
    TASK_DEINDEX_INSTANCE_MAX_RETRIES, TASK_INDEX_INSTANCE_MAX_RETRIES,
    TASK_INDEX_RELATED_INSTANCE_M2M_MAX_RETRIES,
    TASK_INDEX_SEARCH_MODEL_MAX_RETRIES
)
from .settings import setting_indexing_chunk_size

logger = logging.getLogger(name=__name__)


@app.task(
    bind=True, ignore_result=True,
    max_retries=TASK_DEINDEX_INSTANCE_MAX_RETRIES, retry_backoff=True
)
def task_deindex_instance(self, app_label, model_name, object_id):
    logger.info('Executing')

    Model = apps.get_model(app_label=app_label, model_name=model_name)
    instance = Model._meta.default_manager.get(pk=object_id)

    try:
        SearchBackend.get_instance().deindex_instance(instance=instance)
    except (DynamicSearchRetry, LockError) as exception:
        raise self.retry(exc=exception)

    logger.info('Finished')


@app.task(
    bind=True, ignore_result=True,
    max_retries=TASK_INDEX_SEARCH_MODEL_MAX_RETRIES, retry_backoff=True
)
def task_index_search_model(self, search_model_full_name, range_string=None):
    search_model = SearchModel.get(name=search_model_full_name)

    kwargs = {
        'range_string': range_string, 'search_model': search_model
    }

    try:
        SearchBackend.get_instance().index_search_model(**kwargs)
    except (DynamicSearchRetry, LockError) as exception:
        raise self.retry(exc=exception)
    except Exception as exception:
        error_message = (
            'Unexpected error calling `task_index_search_model` with '
            'keyword arguments {}.'
        ).format(kwargs)

        logger.error(error_message)
        raise DynamicSearchException(error_message) from exception


@app.task(
    bind=True, ignore_result=True,
    max_retries=TASK_INDEX_INSTANCE_MAX_RETRIES, retry_backoff=True
)
def task_index_instance(
    self, app_label, model_name, object_id, exclude_app_label=None,
    exclude_model_name=None, exclude_kwargs=None
):
    logger.info('Executing')

    Model = apps.get_model(app_label=app_label, model_name=model_name)
    if exclude_app_label and exclude_model_name:
        ExcludeModel = apps.get_model(
            app_label=exclude_app_label, model_name=exclude_model_name
        )
    else:
        ExcludeModel = None

    try:
        instance = Model._meta.default_manager.get(pk=object_id)
    except Model.DoesNotExist as exception:
        raise self.retry(exc=exception)

    try:
        SearchBackend.get_instance().index_instance(
            instance=instance, exclude_model=ExcludeModel,
            exclude_kwargs=exclude_kwargs
        )
    except (DynamicSearchRetry, LockError) as exception:
        raise self.retry(exc=exception)
    except Exception as exception:
        kwargs = {
            'app_label': app_label,
            'model_name': model_name,
            'object_id': object_id,
            'exclude_app_label': exclude_app_label,
            'exclude_model_name': exclude_model_name,
            'exclude_kwargs': exclude_kwargs
        }
        error_message = (
            'Unexpected error calling `task_index_instance` with keyword '
            'arguments {}.'
        ).format(kwargs)

        logger.error(error_message)
        raise DynamicSearchException(error_message) from exception
    else:
        logger.info('Finished')


@app.task(
    bind=True, ignore_result=True,
    max_retries=TASK_INDEX_RELATED_INSTANCE_M2M_MAX_RETRIES,
    retry_backoff=True
)
def task_index_related_instance_m2m(
    self, action, instance_app_label, instance_model_name,
    instance_object_id, model_app_label, model_model_name, pk_set,
    serialized_search_model_related_paths
):
    InstanceModel = apps.get_model(
        app_label=instance_app_label, model_name=instance_model_name
    )
    instance = InstanceModel.objects.get(pk=instance_object_id)

    Model = apps.get_model(
        app_label=model_app_label, model_name=model_model_name
    )

    search_model_related_paths = {}

    for key, value in serialized_search_model_related_paths.items():
        app_label, model_name = key.split('.')
        DeserializedModel = apps.get_model(
            app_label=app_label, model_name=model_name
        )
        search_model_related_paths[DeserializedModel] = value

    SearchBackend.index_related_instance_m2m(
        action=action, instance=instance, model=Model, pk_set=pk_set,
        search_model_related_paths=search_model_related_paths
    )


@app.task(ignore_result=True)
def task_reindex_backend():
    backend = SearchBackend.get_instance()
    backend.reset()

    for search_model in SearchModel.all():
        queryset = search_model.model._meta.managers_map[search_model.manager_name].all()
        if queryset:
            queryset = queryset.aggregate(min_id=Min('id'), max_id=Max('id'))
            step = setting_indexing_chunk_size.value - 1

            for id_start in range(queryset['min_id'], queryset['max_id'] + 1, step):
                range_string = '{}-{}'.format(id_start, id_start + step - 1)
                task_index_search_model.apply_async(
                    kwargs={
                        'range_string': range_string,
                        'search_model_full_name': search_model.get_full_name(),
                    }
                )



from django.core.management.base import BaseCommand

from mayan.apps.common.utils import parse_range

from ...classes import SearchModel
from ...tasks import task_index_search_model


class Command(BaseCommand):
    help = 'Index a list of objects.'
    missing_args_message = 'You must provide a search model name and an ID range.'

    def add_arguments(self, parser):
        parser.add_argument(
            dest='model_name', help='Name of search model to index.',
            metavar='<model name>'
        )
        parser.add_argument(
            dest='id_range_string', help='Range of IDs to index.',
            metavar='<ID range>'
        )

    def handle(self, model_name, id_range_string, **options):
        try:
            search_model = SearchModel.get(name=model_name)
        except KeyError:
            self.stderr.write(
                msg='Unknown search model `{}`'.format(model_name)
            )
            exit(1)

        try:
            id_range = parse_range(range_string=id_range_string)
        except Exception as exception:
            self.stderr.write(
                msg='Unknown or invalid range format `{}`; {}'.format(
                    id_range_string, exception
                )
            )
            exit(1)

        task_index_search_model.apply_async(
            kwargs={
                'range_string': id_range_string,
                'search_model_full_name': search_model.get_full_name()
            }
        )

        self.stdout.write(
            msg='\nInstances queued for indexing: {}'.format(
                len(list(id_range))
            )
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.converter.permissions import (
    permission_transformation_delete, permission_transformation_edit
)
from mayan.apps.navigation.classes import Link

from ..icons import (
    icon_document_version_active, icon_document_version_create,
    icon_document_version_delete, icon_document_version_edit,
    icon_document_version_export, icon_document_version_list,
    icon_document_version_print, icon_document_version_return_document,
    icon_document_version_return_list, icon_document_version_preview,
    icon_document_version_transformations_clear,
    icon_document_version_transformations_clone
)
from ..permissions import (
    permission_document_version_create, permission_document_version_delete,
    permission_document_version_edit, permission_document_version_export,
    permission_document_version_print, permission_document_version_view,
    permission_document_view
)


link_document_version_active = Link(
    args='resolved_object.pk',
    icon=icon_document_version_active,
    permissions=(permission_document_version_edit,),
    text=_('Make active'), view='documents:document_version_active'
)
link_document_version_create = Link(
    args='resolved_object.pk', icon=icon_document_version_create,
    permissions=(permission_document_version_create,),
    text=_('Create document version'),
    view='documents:document_version_create'
)
link_document_version_delete_single = Link(
    args='resolved_object.pk',
    icon=icon_document_version_delete,
    permissions=(permission_document_version_delete,), tags='dangerous',
    text=_('Delete'), view='documents:document_version_delete_single'
)
link_document_version_delete_multiple = Link(
    icon=icon_document_version_delete, tags='dangerous',
    text=_('Delete'), view='documents:document_version_delete_multiple'
)
link_document_version_edit = Link(
    args='resolved_object.pk', icon=icon_document_version_edit,
    permissions=(permission_document_version_edit,),
    text=_('Edit'), view='documents:document_version_edit'
)
link_document_version_export = Link(
    args='resolved_object.pk', icon=icon_document_version_export,
    permissions=(permission_document_version_export,),
    text=_('Export'), view='documents:document_version_export'
)
link_document_version_list = Link(
    args='resolved_object.pk', icon=icon_document_version_list,
    permissions=(permission_document_version_view,), text=_('Versions'),
    view='documents:document_version_list'
)
link_document_version_preview = Link(
    args='resolved_object.pk', icon=icon_document_version_preview,
    permissions=(permission_document_version_view,),
    text=_('Preview'), view='documents:document_version_preview'
)
link_document_version_print_form = Link(
    args='resolved_object.id', icon=icon_document_version_print,
    permissions=(permission_document_version_print,), text=_('Print'),
    view='documents:document_version_print_form'
)
link_document_version_return_to_document = Link(
    args='resolved_object.document.pk',
    icon=icon_document_version_return_document,
    permissions=(permission_document_view,), text=_('Document'),
    view='documents:document_preview'
)
link_document_version_return_list = Link(
    args='resolved_object.document.pk',
    icon=icon_document_version_return_list,
    permissions=(permission_document_version_view,), text=_('Versions'),
    view='documents:document_version_list'
)
link_document_version_transformations_clear = Link(
    args='resolved_object.id',
    icon=icon_document_version_transformations_clear,
    permissions=(permission_transformation_delete,),
    text=_('Clear transformations'),
    view='documents:document_version_transformations_clear'
)
link_document_version_multiple_transformations_clear = Link(
    icon=icon_document_version_transformations_clear,
    permissions=(permission_transformation_delete,),
    text=_('Clear transformations'),
    view='documents:document_version_multiple_transformations_clear'
)
link_document_version_transformations_clone = Link(
    args='resolved_object.id',
    icon=icon_document_version_transformations_clone,
    permissions=(permission_transformation_edit,),
    text=_('Clone transformations'),
    view='documents:document_version_transformations_clone'
)



from mayan.apps.dynamic_search.tests.mixins import SearchTestMixin

from ..permissions import permission_document_version_view
from ..search import document_version_page_search, document_version_search

from .base import GenericDocumentViewTestCase
from .literals import TEST_DOCUMENT_VERSION_COMMENT


class DocumentVersionSearchTestCase(
    SearchTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def _do_test_search(self, query):
        terms = str(tuple((query.values()))[0]).strip()
        self.assertTrue(terms is not None)
        self.assertTrue(terms != '')

        return self.search_backend.search(
            search_model=document_version_search, query=query,
            user=self._test_case_user
        )

    def setUp(self):
        super().setUp()
        self._upload_test_document(
            document_version_attributes={
                'comment': TEST_DOCUMENT_VERSION_COMMENT
            }
        )

    def test_document_version_search_by_comment_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'comment': self.test_document_version.comment
            }
        )
        self.assertTrue(self.test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_search_by_comment_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'comment': self.test_document_version.comment
            }
        )
        self.assertTrue(self.test_document_version in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_search_by_comment_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'comment': self.test_document_version.comment
            }
        )
        self.assertTrue(self.test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_search_by_document_description_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__description': self.test_document.description
            }
        )
        self.assertTrue(self.test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_search_by_document_description_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__description': self.test_document.description
            }
        )
        self.assertTrue(self.test_document_version in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_search_by_document_description_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__description': self.test_document.description
            }
        )
        self.assertTrue(self.test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_search_by_document_label_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__label': self.test_document.label
            }
        )
        self.assertTrue(self.test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_search_by_document_label_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__label': self.test_document.label
            }
        )
        self.assertTrue(self.test_document_version in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_search_by_document_label_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__label': self.test_document.label
            }
        )
        self.assertTrue(self.test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_search_by_document_uuid_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__uuid': self.test_document.uuid
            }
        )
        self.assertTrue(self.test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_search_by_document_uuid_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__uuid': self.test_document.uuid
            }
        )
        self.assertTrue(self.test_document_version in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_search_by_document_uuid_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__uuid': self.test_document.uuid
            }
        )
        self.assertTrue(self.test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_search_by_document_type_label_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__document_type__label': self.test_document_type.label
            }
        )
        self.assertTrue(self.test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_search_by_document_type_label_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__document_type__label': self.test_document_type.label
            }
        )
        self.assertTrue(self.test_document_version in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_search_by_document_type_label_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document__document_type__label': self.test_document_type.label
            }
        )
        self.assertTrue(self.test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentVersionPageSearchTestCase(
    SearchTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def _do_test_search(self, query):
        terms = str(tuple((query.values()))[0]).strip()
        self.assertTrue(terms is not None)
        self.assertTrue(terms != '')

        return self.search_backend.search(
            search_model=document_version_page_search, query=query,
            user=self._test_case_user
        )

    def setUp(self):
        super().setUp()
        self._upload_test_document(
            document_version_attributes={
                'comment': TEST_DOCUMENT_VERSION_COMMENT
            }
        )

    def test_document_version_page_search_by_document_label_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__label': self.test_document.label
            }
        )
        self.assertTrue(self.test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_search_by_document_label_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__label': self.test_document.label
            }
        )
        self.assertTrue(self.test_document_version_page in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_search_by_document_label_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__label': self.test_document.label
            }
        )
        self.assertTrue(self.test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_search_by_document_description_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__description': self.test_document.description
            }
        )
        self.assertTrue(self.test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_search_by_document_description_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__description': self.test_document.description
            }
        )
        self.assertTrue(self.test_document_version_page in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_search_by_document_description_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__description': self.test_document.description
            }
        )
        self.assertTrue(self.test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_search_by_document_uuid_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__uuid': self.test_document.uuid
            }
        )
        self.assertTrue(self.test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_search_by_document_uuid_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__uuid': self.test_document.uuid
            }
        )
        self.assertTrue(self.test_document_version_page in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_search_by_document_uuid_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__uuid': self.test_document.uuid
            }
        )
        self.assertTrue(self.test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_search_by_document_type_label_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__document_type__label': self.test_document_type.label
            }
        )
        self.assertTrue(self.test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_search_by_document_type_label_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__document_type__label': self.test_document_type.label
            }
        )
        self.assertTrue(self.test_document_version_page in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_search_by_document_type_label_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__document__document_type__label': self.test_document_type.label
            }
        )
        self.assertTrue(self.test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_search_by_document_version_comment_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__comment': self.test_document_version.comment
            }
        )
        self.assertTrue(self.test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_page_search_by_document_version_comment_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__comment': self.test_document_version.comment
            }
        )
        self.assertTrue(self.test_document_version_page in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_version_page_search_by_document_version_comment_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_version_view
        )

        self.test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'document_version__comment': self.test_document_version.comment
            }
        )
        self.assertTrue(self.test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)
